# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import pymatgen.analysis.structure_matcher as sm
from pymatgen.core.structure import Molecule
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.molecule import pyxtal_molecule
from pyxtal.symmetry import Group, Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestMolecule(unittest.TestCase):
    def test_get_orientations_in_wp(self):
        m = pyxtal_molecule("Benzene")
        g = Group(61)
        assert len(m.get_orientations_in_wp(g[0])) == 1
        assert len(m.get_orientations_in_wp(g[1])) == 1
        assert len(m.get_orientations_in_wp(g[2])) == 1


class TestMolecular(unittest.TestCase):
    def test_single_specie(self):
        struc = pyxtal(molecular=True)
        struc.from_random(3, 36, ["H2O"], sites=[["8b"]])
        struc.to_file("tmp-molecular.cif")
        os.remove("tmp-molecular.cif")
        assert struc.valid

        # test space group
        pmg_struc = struc.to_pymatgen()
        sga = SpacegroupAnalyzer(pmg_struc)
        # print(sga.get_space_group_symbol())
        assert sga.get_space_group_number() >= 36

        # test rotation
        struc.mol_sites[0].orientation.axis
        struc.mol_sites[0].rotate(ax_vector=[1, 0, 0], angle=90)
        pmg_struc = struc.to_pymatgen()
        sga = SpacegroupAnalyzer(pmg_struc)
        # pmg_struc.to("cif", "1.cif")
        assert sga.get_space_group_symbol() == "Cmc2_1"
        # print(pmg_struc.frac_coords[:3])

    def test_sites(self):
        struc = pyxtal(molecular=True)
        struc.from_random(3, 19, ["H2O"])
        pmg_struc = struc.to_pymatgen()
        sga = SpacegroupAnalyzer(pmg_struc)
        assert sga.get_space_group_symbol() == "P2_12_12_1"

        struc = pyxtal(molecular=True)
        struc.from_random(3, 36, ["H2O"], [8], sites=[["4a", "4a"]])
        pmg_struc = struc.to_pymatgen()
        sga = SpacegroupAnalyzer(pmg_struc)
        assert sga.get_space_group_symbol() == "Cmc2_1"

    def test_sites_xyz(self):
        struc = pyxtal(molecular=True)
        sites = [{"4e": [0.77, 0.57, 0.53]}]
        lat = Lattice.from_para(11.43, 6.49, 11.19, 90, 83.31, 90, ltype="monoclinic")
        struc.from_random(3, 14, ["aspirin"], [4], lattice=lat, sites=sites)
        assert struc.valid

    def test_read(self):
        # test reading structure from external
        struc = pyxtal(molecular=True)
        struc.from_seed(seed=cif_path + "aspirin.cif", molecules=["aspirin"])
        assert struc.lattice.ltype == "monoclinic"
        pmg_struc = struc.to_pymatgen()
        sga = SpacegroupAnalyzer(pmg_struc)
        assert sga.get_space_group_symbol() == "P2_1/c"
        C = struc.subgroup_once(eps=0, H=4)
        pmg_s2 = C.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg_struc, pmg_s2)

    def test_big_molecule(self):
        # print("test_big_molecule")
        for mol in ["ROY", "aspirin"]:
            struc = pyxtal(molecular=True)
            struc.from_random(3, 19, [mol], factor=1.4)
            assert struc.valid
            pair = struc.check_short_distances()
            if len(pair) > 0:
                print("short distances were detected")
                print(mol)
                print(pair)
            assert len(pair) == 0

    def test_from_random_site(self):
        spg, wps, elements, numIons = 224, [["24j"]], ["C"], [24]
        for _i in range(10):
            c = pyxtal()
            c.from_random(3, spg, elements, numIons, sites=wps)
            pair = c.check_short_distances(0.5)
            assert len(pair) == 0

    def test_c60(self):
        struc = pyxtal(molecular=True)
        struc.from_random(3, 36, ["C60"], [4], 1.0)
        assert struc.valid

    def test_mutiple_species(self):
        Li = Molecule(["Li"], [[0.0, 0.0, 0.0]])
        coords = [
            [0.000000, 0.000000, 0.000000],
            [1.200000, 1.200000, -1.200000],
            [1.200000, -1.200000, 1.200000],
            [-1.200000, 1.200000, 1.200000],
            [-1.200000, -1.200000, -1.200000],
        ]
        ps4 = Molecule(["P", "S", "S", "S", "S"], coords)

        for _i in range(3):
            struc = pyxtal(molecular=True)
            struc.from_random(3, 10, [Li, ps4], [6, 2], 1.2, conventional=False)
            if struc.valid:
                assert len(struc.to_pymatgen()) == 16

    def test_distance(self):
        Ag_xyz = """1
        AgC2N2H
        Ag         4.30800        8.26300       -0.2200
        """

        C2N2H7_xyz = """12
        AgC2N2H
        H          5.95800        5.80600       -0.9530
        N          5.24100        6.16800       -1.1210
        N          2.23200        6.99000       -0.6820
        C          4.02900        5.47000       -1.5870
        C          2.78500        5.61100       -0.6610
        H          3.69300        5.63500       -2.5830
        H          4.17400        4.42800       -1.6990
        H          3.12700        5.50500        0.4260
        H          2.05000        5.01200       -0.9300
        H          1.96000        7.20500       -1.3860
        H          1.59400        6.99200       -0.0710
        """
        with open("Ag.xyz", "w") as f:
            f.write(Ag_xyz)
        with open("C2N2H7.xyz", "w") as f:
            f.write(C2N2H7_xyz)

        for _i in range(10):
            c = pyxtal(molecular=True)
            c.from_random(3, 9, ["Ag.xyz", "C2N2H7.xyz"], [12, 12])
            short_bonds = c.check_short_distances(r=1.1)
            assert len(short_bonds) == 0
        os.remove("Ag.xyz")
        os.remove("C2N2H7.xyz")

    def test_molecular_2d(self):
        # print("test_molecular_2d")
        struc = pyxtal(molecular=True)
        struc.from_random(2, 20, ["H2O"])
        assert struc.valid

    def test_molecular_1d(self):
        struc = pyxtal(molecular=True)
        struc.from_random(1, 20, ["H2O"])
        assert struc.valid

    def test_preassigned_sites(self):
        sites = [["4a", "4a"]]
        struc = pyxtal(molecular=True)
        struc.from_random(3, 36, ["H2O"], [8], sites=sites)
        assert struc.valid

    def test_special_sites(self):
        struc = pyxtal(molecular=True)
        struc.from_random(3, 61, ["Benzene"], [4])
        assert struc.valid

if __name__ == "__main__":
    unittest.main()



"""
Module for generating and analyzing transformation operations. Several functions
for working with matrices are provided. The class OperationAnalyzer allows for
comparison between pymatgen.core.operations.SymmOp objects, and can be used to
identify conjugate operations. The orientation class can be used to identify
degrees of freedom for molecules in Wyckoff positions with certain symmetry
constraints.
"""

from copy import deepcopy

import numpy as np
from pymatgen.core import Element
from pymatgen.core.operations import SymmOp
from scipy.spatial.distance import cdist
from scipy.spatial.transform import Rotation

from pyxtal.constants import all_sym_directions, deg, hex_cell, rad
from pyxtal.tolerance import Tol_matrix


# ------------------------------
# Define functions
def check_distance(
    coord1,
    coord2,
    species1,
    species2,
    lattice,
    PBC=None,
    tm: Tol_matrix = None,
    d_factor=1.0,
):
    """
    Check the distances between two atom set. Only distances between points from
    different sets are checked.

    Args:
        coord1: a list of fractional coordinates e.g. [[.1,.6,.4]
            [.3,.8,.2]]
        coord2: a list of new fractional coordinates e.g. [[.7,.8,.9],
            [.4,.5,.6]]
        species1: a list of atomic species or numbers for coord1
        species2: a list of atomic species or numbers for coord2
        lattice: matrix describing the unit cell vectors
        PBC: A periodic boundary condition list,
            where 1 means periodic, 0 means not periodic.
            [1,1,1] -> full 3d periodicity,
            [0,0,1] -> periodicity along the z axis
        tm: a Tol_matrix object, or a string representing Tol_matrix
        d_factor: the tolerance is multiplied by this amount. Larger values
            mean atoms must be farther apart

    Returns:
        a bool for whether or not the atoms are sufficiently far enough apart
    """
    if tm is None:
        tm = Tol_matrix(prototype="atomic")

    # Check that there are points to compare
    if PBC is None:
        PBC = [1, 1, 1]
    if len(coord1) < 1 or len(coord2) < 1:
        return True

    # Create tolerance matrix from subset of tm
    tols = np.zeros((len(species1), len(species2)))
    for i1, specie1 in enumerate(species1):
        for i2, specie2 in enumerate(species2):
            tols[i1][i2] = tm.get_tol(specie1, specie2)

    # Calculate the distance between each i, j pair
    d = distance_matrix(coord1, coord2, lattice, PBC=PBC)

    return not (np.array(d) < np.array(tols)).any()


def verify_distances(coordinates, species, lattice, factor=1.0, PBC=None):
    """
    Checks the inter-atomic distance between all pairs of atoms in a crystal.

    Args:
        coordinates: a 1x3 list of fractional coordinates
        species: a list of atomic symbols for each coordinate
        lattice: a 3x3 matrix representing the lattice vectors of the unit cell
        factor: a tolerance factor for checking distances. A larger value means
            atoms must be farther apart
        PBC: A periodic boundary condition list, where 1 means periodic, 0 means
            not periodic. Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> 1d
            periodicity along the z axis

    Returns:
        True if no atoms are too close together, False if any pair is too close
    """
    if PBC is None:
        PBC = [1, 1, 1]
    for i, c1 in enumerate(coordinates):
        specie1 = species[i]
        for j, c2 in enumerate(coordinates):
            if j > i:
                specie2 = species[j]
                diff = np.array(c2) - np.array(c1)
                d_min = distance(diff, lattice, PBC=PBC)
                radius = Element(specie1).covalent_radius + Element(specie2).covalent_radius
                tol = factor * 0.5 * radius
                if d_min < tol:
                    return False
    return True


def check_images(
    coords,
    species,
    lattice,
    PBC=None,
    tm: Tol_matrix = None,
    tol=None,
    d_factor=1.0,
):
    """
    Given a set of (unfiltered) frac coordinates, checks if the periodic images
    are too close.

    Args:
        coords: a list of fractional coordinates
        species: the atomic species of each coordinate
        lattice: a 3x3 lattice matrix
        PBC: the periodic boundary conditions
        tm: a Tol_matrix object
        tol: a single override value for the distance tolerances
        d_factor: the tolerance is multiplied by this amount. Larger values
            mean atoms must be farther apart

    Returns:
        False if distances are too close. True if distances are not too close
    """
    if tm is None:
        tm = Tol_matrix(prototype="atomic")

    # If no PBC, there are no images to check
    if PBC is None:
        PBC = [1, 1, 1]

    if PBC == [0, 0, 0]:
        return True

    # Create image coords from given coords and PBC
    coords = np.array(coords)
    m = create_matrix(PBC=PBC, omit=True)

    new_coords = []
    for v in m:
        for v2 in coords + v:
            new_coords.append(v2)

    new_coords = np.array(new_coords)
    # Create a distance matrix
    dm = distance_matrix(coords, new_coords, lattice, PBC=[0, 0, 0])
    # Define tolerances
    if tol is None:
        tols = np.zeros((len(species), len(species)))
        for i, s1 in enumerate(species):
            for j, s2 in enumerate(species):
                if i <= j:
                    tols[i][j] = tm.get_tol(s1, s2)
                    tols[j][i] = tm.get_tol(s1, s2)
        tols2 = np.tile(tols, int(len(new_coords) / len(coords)))
        return not (dm < tols2).any()
    elif tol is not None:
        return not (dm < tol).any()
    return True


def distance(xyz, lattice, PBC=None):
    """
    Returns the Euclidean distance from the origin for a fractional
    displacement vector. Takes into account the lattice metric and periodic
    boundary conditions, including up to one non-periodic axis.

    Args:
        xyz: a fractional 3d displacement vector. Can be obtained by
            subtracting one fractional vector from another
        lattice: a 3x3 matrix describing a unit cell's lattice vectors
        PBC: A periodic boundary condition list, where 1 means periodic, 0 means
        not periodic. Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> 1d
        periodicity along the z axis

    Returns:
        a scalar for the distance of the point from the origin
    """
    if PBC is None:
        PBC = [1, 1, 1]
    xyz = filtered_coords(xyz, PBC=PBC)
    matrix = create_matrix(PBC=PBC)
    matrix += xyz
    matrix = np.dot(matrix, lattice)
    return np.min(np.linalg.norm(matrix, axis=1))


def distance_matrix(pts1, pts2, lattice, PBC=None, single=False, metric="euclidean"):
    """
    Returns the distances between two sets of fractional coordinates.
    Takes into account the lattice metric and periodic boundary conditions.

    Args:
        pts1: a list of fractional coordinates
        pts2: another list of fractional coordinates
        lattice: a 3x3 matrix describing a unit cell's lattice vectors
        PBC: A periodic boundary condition list, where 1 means periodic, 0 means
            not periodic. Ex: [1,1,1] -> full 3d periodicity, [0,0,1] -> 1d
            periodicity along the z axis
        single: return a scalor and matrix?
        metric: the metric to use with cdist, such as  'euclidean',
            'sqeuclidean', 'minkowski', and others

    Returns:
        a scalor or distance matrix
    """
    if PBC is None:
        PBC = [1, 1, 1]

    if PBC != [0, 0, 0]:
        l1 = filtered_coords(pts1, PBC=PBC)
        l2 = filtered_coords(pts2, PBC=PBC)
        l1 = np.dot(l1, lattice)
        l2 = np.dot(l2, lattice)
        matrix = create_matrix(PBC=PBC)
        matrix = np.dot(matrix, lattice)
        all_distances = np.zeros([len(matrix), len(l1), len(l2)])
        for i, v in enumerate(matrix):
            all_distances[i] += cdist(l1 + v, l2, metric)
        # m1 = np.array([(l1 + v) for v in matrix])
        # m1 = np.vstack([l1 + v for v in matrix])
        # all_distances = np.array([cdist(l, l2, metric) for l in m1])
        if single:
            return np.min(all_distances)
        else:
            # return np.apply_along_axis(np.min, 0, all_distances)
            return np.min(all_distances, axis=0)

    else:
        return distance_matrix_no_PBC(pts1, pts2, lattice, single, metric)


def distance_matrix_no_PBC(pts1, pts2, lattice, single=False, metric="euclidean"):
    """
    Returns the distances between two sets of fractional coordinates.
    Without periodic boundary conditions.

    Args:
        pts1: a list of fractional coordinates (N1*3)
        pts2: another list of fractional coordinates (N2*3)
        lattice: a 3x3 matrix describing a unit cell's lattice vectors
        single: return the minimum distance or the matrix
        metric: the metric to use with cdist. e.g. `euclidean`,
            `sqeuclidean`, `minkowski`, and others

    Returns:
        a scalor or distance matrix
    """

    l1 = np.dot(pts1, lattice)
    l2 = np.dot(pts2, lattice)
    d = cdist(l1, l2, metric)
    if single:
        return np.min(d)
    else:
        return d


def create_matrix(PBC=None, omit=False):
    """
    Used for calculating distances in lattices with periodic boundary
    conditions. When multiplied with a set of points, generates additional
    points in cells adjacent to and diagonal to the original cell

    Args:
        PBC: A periodic boundary condition list (1: periodic; 0: nonperiodic).

    Returns:
        A numpy array which can be multiplied by a set of coordinates
    """
    if PBC is None:
        PBC = [1, 1, 1]
    matrix = []
    i_list = [-1, 0, 1] if PBC[0] else [0]
    j_list = [-1, 0, 1] if PBC[1] else [0]
    k_list = [-1, 0, 1] if PBC[2] else [0]
    for i in i_list:
        for j in j_list:
            for k in k_list:
                if omit:
                    if [i, j, k] != [0, 0, 0]:
                        matrix.append([i, j, k])
                else:
                    matrix.append([i, j, k])
    return np.array(matrix, dtype=float)


def filtered_coords(coords, PBC=None):
    """
    Transform all coordinates to [0, 1] interval if PBC is allowed
    For example, [1.2, 1.6, -.4] becomes
    [0.2, 0.6, 0.6] when PBC=[1,1,1]
    [0.2, 1.6, 0.6] when PBC=[1,0,1]

    Args:
        coords: an array of real 3d vectors.
        PBC: A periodic boundary condition list (1: periodic; 0: nonperiodic).

    Returns:
        an array of filtered coords with the same shape as coords
    """

    if PBC is None:
        PBC = [1, 1, 1]
    if isinstance(coords, list):
        coords = np.array(coords)
    for i in range(3):
        if PBC[i] > 0:
            if len(coords.shape) > 1:
                coords[:, i] -= np.floor(coords[:, i])
            else:
                coords[i] -= np.floor(coords[i])
    return coords


def filtered_coords_euclidean(coords, PBC=None):
    """
    Given an array of fractional 3-vectors, filters coordinates to between 0 and
    1. Then, values which are greater than 0.5 are converted to 1 minus their
    value. This is used for converting displacement vectors with a Euclidean
    lattice.

    Args:
        coords: an array of real 3d vectors. The shape does not matter
        PBC: A periodic boundary condition list (1: periodic; 0: nonperiodic).

    Returns:
        an array of filtered coords with the same shape as coords
    """

    if PBC is None:
        PBC = [1, 1, 1]

    def filter_vector_euclidean(vector):
        for i, a in enumerate(PBC):
            if a:
                # QZ: check if this is equivalent to -= np.rint()
                vector[i] -= np.floor(vector[i])
                if vector[i] > 0.5:
                    vector[i] = 1 - vector[i]
        return vector

    return np.apply_along_axis(filter_vector_euclidean, -1, coords)


def get_inverse(op):
    """
    Given a SymmOp object, returns its inverse.

    Args:
        op: a Symmop object

    Returns:
        the inverse
    """
    matrix = op.affine_matrix.copy()
    # fill the matrix if it is ill conditioned
    # experimental
    if np.linalg.matrix_rank(matrix) < 4:
        for row in range(3):
            # fixed value
            if np.sum(matrix[row, :3] ** 2) < 1e-3:
                matrix[row, row] = 1
                matrix[row, 3] = 0

        if np.linalg.matrix_rank(matrix) == 3:
            # [-3x/2, -x/2, 1/4]
            # [0, x, 1/4]
            for rows in [[0, 1, 2], [1, 2, 0], [0, 2, 1]]:
                # m = (matrix[rows,:])[:,rows]
                # print(rows, m)
                if np.linalg.matrix_rank(matrix[rows[:2], :3]) != 2:
                    break
            id0, id1, id2 = rows[0], rows[1], rows[2]
            if matrix[id0, id1] == 0:
                matrix[id0, id1], matrix[id0, id0] = matrix[id0, id0], matrix[id0, id1]
                if np.linalg.matrix_rank(matrix) == 3:
                    matrix[id0, id1], matrix[id0, id2] = (
                        matrix[id0, id2],
                        matrix[id0, id1],
                    )
            else:
                matrix[id1, id0], matrix[id1, id1] = matrix[id1, id1], matrix[id1, id0]
                if np.linalg.matrix_rank(matrix) == 3:
                    matrix[id1, id0], matrix[id1, id2] = (
                        matrix[id1, id2],
                        matrix[id1, id0],
                    )

        elif np.linalg.matrix_rank(matrix) == 2:
            # -3x/2, -x/2, -x+1/4
            if np.sum(matrix[:, 0] ** 2) > 1e-3:
                matrix[1, 0], matrix[1, 1] = matrix[1, 1], matrix[1, 0]
                matrix[2, 0], matrix[2, 2] = matrix[2, 2], matrix[2, 0]
            elif np.sum(matrix[:, 1] ** 2) > 1e-3:
                matrix[0, 1], matrix[0, 0] = matrix[0, 0], matrix[0, 1]
                matrix[2, 1], matrix[2, 2] = matrix[2, 2], matrix[2, 1]
            else:
                matrix[0, 2], matrix[0, 0] = matrix[0, 0], matrix[0, 2]
                matrix[1, 2], matrix[1, 1] = matrix[1, 1], matrix[1, 2]
    return SymmOp(np.linalg.inv(matrix))


def get_inverse_ops(ops):
    """
    Given a inverse list of SymmOp objects

    Args:
        ops: a list of Symmop's

    Returns:
        a list of equal shape to ops, with the inverse operations
    """
    inverses = []
    for op in ops:
        if type(op) == SymmOp:
            inverses.append(op.inverse)
        else:
            inverses.append(get_inverse_ops(op))
    return inverses


def apply_ops(coord, ops):
    """
    Apply a list of SymmOps to a single 3-vector and return an array of
    the generated vectors. This is the inverse of SymmOp.operate_multi.

    Args:
        coord: a 3-vector (list or numpy array)
        ops: a list, tuple, or array of SymmOp objects

    Returns:
        an np array of floating-point 3-vectors
    """
    coord = np.array(coord)
    affine_point = np.concatenate([coord, np.ones(coord.shape[:-1] + (1,))], axis=-1)
    matrices = np.array([op.affine_matrix for op in ops])
    return np.inner(affine_point, matrices)[..., :-1]


def apply_ops_diagonal(coords, ops):
    """
    Given a list of coordinates and SymmOps, apply the ith op to the ith coord
    and return the list of transformed coordinates

    Args:
        coords: a list or array of 3-vectors

    Returns:
        a transformed numpy array of 3-vectors
    """
    coords = np.array(coords)
    affine_points = np.concatenate([coords, np.ones(coords.shape[:-1] + (1,))], axis=-1)
    matrices = np.array([op.affine_matrix for op in ops])
    return np.einsum("...ij,...j", matrices, affine_points)[:, :3]


def angle(v1, v2, radians=True):
    """
    Calculate the angle (in radians) between two vectors.

    Args:
        v1: a 1x3 vector
        v2: a 1x3 vector
        radians: whether to return angle in radians (default) or degrees

    Returns:
        the angle in radians between the two vectors
    """
    v1 = np.real(v1)
    v2 = np.real(v2)
    dot = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
    # if np.isclose(dot, 1.0):
    if np.abs(dot - 1) < 1e-3:
        a = 0
    elif np.abs(dot + 1) < 1e-3:
        a = np.pi
    else:
        a = np.arccos(dot)

    if radians:
        return a
    else:
        return a * deg


def is_orthogonal(m, tol=0.001):
    """
    Check whether or not a 3x3 matrix is orthogonal. An orthogonal matrix has
    the property that it is equal to its transpose.

    Args:
        m: a 3x3 matrix (list or numpy array)
        tol: the numerical tolerance for checking if two matrices are equal

    Returns:
        True if the matrix is orthogonal, False if it is not
    """
    m1 = np.dot(m, np.transpose(m))
    m2 = np.dot(np.transpose(m), m)
    return not (not np.allclose(m1, np.identity(3), rtol=tol) or not np.allclose(m2, np.identity(3), rtol=tol))


def aa2matrix(axis, angle, radians=True, random=False):
    """
    Given an axis and an angle, return a 3x3 rotation matrix.
    Based on:
    https://en.wikipedia.org/wiki/Rotation_matrix#Axis_and_angle

    Args:
        axis: a vector about which to perform a rotation
        angle: the angle of rotation
        radians: whether the supplied angle is in radians (True)
            or in degrees (False)
        random: whether or not to choose a random rotation matrix. If True, the
            axis and angle are ignored, and a random orientation is generated

    Returns:
        a 3x3 numpy array representing a rotation matrix
    """
    # Convert to radians if necessary
    if not radians:
        angle *= rad
    # Allow for generation of random rotations
    if random:
        # a = np.random.random()
        axis = np.random.sample(3)
        angle = np.random.random() * np.pi * 2
    # Ensure axis is a unit vector
    axis = axis / np.linalg.norm(axis)
    # Define quantities which are reused
    x = np.real(axis[0])
    y = np.real(axis[1])
    z = np.real(axis[2])
    c = np.cos(angle)
    s = np.sin(angle)
    C = 1 - c
    # Define the rotation matrix
    Q = np.zeros([3, 3])
    Q[0][0] = x * x * C + c
    Q[0][1] = x * y * C - z * s
    Q[0][2] = x * z * C + y * s
    Q[1][0] = y * x * C + z * s
    Q[1][1] = y * y * C + c
    Q[1][2] = y * z * C - x * s
    Q[2][0] = z * x * C - y * s
    Q[2][1] = z * y * C + x * s
    Q[2][2] = z * z * C + c
    return Q


def rotate_vector(v1, v2, rtol=1e-4):
    # TODO: Verify that multiplication order is correct
    # (matrix should come after vector in np.dot)
    """
    Rotates a vector v1 to v2 about an axis perpendicular to both. Returns the
    3x3 rotation matrix used to do so.

    Args:
        v1: a 1x3 vector (list or array) of floats
        v2: a 1x3 vector (list or array) of floats

    Returns:
        a 3x3 matrix corresponding to a rotation which
        can be applied to v1 to obtain v2
    """
    v1 = v1 / np.linalg.norm(v1)
    v2 = v2 / np.linalg.norm(v2)
    dot = np.dot(v1, v2)
    # Handle collinear vectors
    if np.abs(dot - 1) < rtol:
        return np.identity(3)
    elif np.abs(dot + 1) < rtol:
        r = np.random.sample(3) #[np.random.random(), np.random.random(), np.random.random()]
        v3 = np.cross(v1, r)
        v3 /= np.linalg.norm(v3)
        # return aa2matrix(v3, np.pi)
        return Rotation.from_rotvec(np.pi * v3).as_matrix()
    theta = angle(v1, v2)
    v3 = np.cross(v1, v2)
    v3 /= np.linalg.norm(v3)
    # return aa2matrix(v3, theta)
    return Rotation.from_rotvec(theta * v3).as_matrix()


def are_equal(op1, op2, PBC=None, rtol=1e-3, atol=1e-3):
    """
    Check whether two SymmOp objects are equal up to some numerical tolerance.
    Allows for optional consideration of periodic boundary conditions. This
    option is useful for handling the periodicity of crystals.

    Args:
        op1: a SymmOp object
        op2: another SymmOp object
        PBC: A periodic boundary condition list.
        rtol: the relative numerical tolerance for equivalence
        atol: the absolute numerical tolerance for equivalence

    Returns:
        True if op1 and op2 are equivalent, False otherwise
    """
    # Check two SymmOps for equivalence
    # pbc=True means integer translations will be ignored
    if PBC is None:
        PBC = [1, 1, 1]
    m1 = op1.rotation_matrix
    m2 = op2.rotation_matrix
    # Check that rotations are equivalent
    if not np.allclose(m1, m2, rtol=rtol, atol=atol):
        return False
    v1 = op1.translation_vector
    v2 = op2.translation_vector

    difference = v2 - v1

    for i, a in enumerate(PBC):
        if a:
            difference[i] -= np.rint(difference[i])

    d = np.linalg.norm(difference)

    return abs(d) < rtol


class OperationAnalyzer(SymmOp):
    """
    Class for comparing operations. Stores rotation axis and angle, as well as
    the type and order of operation (identity, inversion, rotation, or
    rotoinversion). By default, takes a SymmOp as argument. This information can
    be accessed by calling print(object). The class methods is_conjugate and
    are_conjugate can be used to determine if two operations are conjugate to
    each other. That is, whether they represent the same angle of rotation and
    are either both inverted or both uninverted.

    Note: rotoinversions with odd-order rotational parts will have an over-all
        even order. For example, the order of (-3) is 6.

    Note: reflections are treated as rotoinversions of order 2.

    Args:
        SymmOp: a pymatgen.core.structure.SymmOp object to analyze
    """

    # ++++: include support for off-center operations
    # TODO: include support for shear and scaling operations
    # TODO: include support for matrix-column and axis-angle initialization
    def get_order(angle, rotoinversion=False, tol=1e-2):
        # Find the order of a rotation based on its angle
        found = False
        for n in range(1, 61):
            x = (n * angle) / (2.0 * np.pi)
            y = x - np.rint(x)
            if abs(y) <= tol:
                found = True
                break
        if found:
            # Double order of odd-rotation rotoinversions
            if rotoinversion:
                if n % 2 == 1:
                    return int(n * 2)
                else:
                    return int(n)
            else:
                return int(n)
        if not found:
            return "irrational"
        return None

    def __init__(self, op, parse_trans=False, hexagonal=False):
        if type(op) == deepcopy(SymmOp):
            # The numerical tolerance associated with op
            # The 4x4 affine matrix of the op
            # The 3x3 rotation (or rotoinversion) matrix
            # The determinant of self.m
            self.op = op
            self.tol = op.tol
            self.affine_matrix = op.affine_matrix
            self.m = op.rotation_matrix
            self.det = np.linalg.det(self.m)
            self.hexagonal = hexagonal

        elif (type(op) == np.ndarray) or (type(op) == np.matrix):
            if op.shape == (3, 3):
                self.op = SymmOp.from_rotation_and_translation(op, [0, 0, 0])
                self.m = self.op.rotation_matrix
                self.det = np.linalg.det(op)
        else:
            raise ValueError("Error: OperationAnalyzer requires a SymmOp or 3x3 array.")

        self.symbol = None
        self.parse_trans = parse_trans  # only for space group

        # If rotation matrix is not orthogonal
        if not is_orthogonal(self.m):
            self.type = "general"
            self.axis, self.angle, self.order, self.rotation_order = (
                None,
                None,
                None,
                None,
            )
        # If rotation matrix is orthogonal
        else:
            # If determinant is positive, rotation
            if np.linalg.det(self.m) > 0:
                self.inverted = False
                rotvec = Rotation.from_matrix(self.m).as_rotvec()
                if np.sum(rotvec.dot(rotvec)) < 1e-6:
                    self.axis = None
                    self.angle = 0
                else:
                    self.angle = np.linalg.norm(rotvec)
                    self.axis = rotvec / self.angle
                    if self.hexagonal:
                        # print('convert hex', self.axis, np.dot(self.axis, hex_cell))
                        if np.linalg.norm(self.axis - np.array([1, 0, 0])) > 1e-4:
                            self.axis = np.dot(self.axis, hex_cell)
                    # parse symmetry direction
                    if self.parse_trans and not self.parse_axis():
                        self.axis *= -1
                        self.angle = 2 * np.pi - self.angle
                        # print('switch angle', self.angle)

                if np.isclose(self.angle, 0):
                    # Types: 'identity', 'inversion', 'rotation', or 'rotoinversion'.
                    # order: the number of times to get to origin
                    # rotation_order: 2, 3, 4, 6
                    self.type = "identity"
                    self.order = 1
                    self.rotation_order = 1
                    self.symbol = "1"
                else:
                    self.type = "rotation"
                    self.order = OperationAnalyzer.get_order(self.angle)
                    self.rotation_order = self.order
                    if self.parse_trans:
                        self.symbol = self.parse_screw_symmetry()

            # If determinant is negative, rotoinversion (including reflection)
            elif np.linalg.det(self.m) < 0:
                self.inverted = True
                rotvec = Rotation.from_matrix(-1 * self.m).as_rotvec()
                if np.sum(rotvec.dot(rotvec)) < 1e-6:
                    self.axis = None
                    self.angle = 0
                else:
                    self.angle = np.linalg.norm(rotvec)
                    self.axis = rotvec / self.angle
                    if self.hexagonal:
                        if np.linalg.norm(self.axis - np.array([1, 0, 0])) > 1e-4:
                            # print('convert hex', self.axis, np.dot(self.axis, hex_cell))
                            self.axis = np.dot(self.axis, hex_cell)
                if np.isclose(self.angle, 0):
                    self.symbol = "-1"
                    self.type = "inversion"
                    self.order = 2
                    self.rotation_order = 1
                else:
                    # parse symmetry direction
                    # if self.parse_trans and not self.parse_axis():
                    #    self.axis *= -1
                    #    self.angle = 2*np.pi - self.angle
                    #    print('switch angle', self.angle)

                    self.axis *= -1
                    self.type = "rotoinversion"
                    self.order = OperationAnalyzer.get_order(self.angle, rotoinversion=True)
                    self.rotation_order = OperationAnalyzer.get_order(self.angle, rotoinversion=False)
                    if self.parse_trans:
                        self.symbol = self.parse_glide_symmetry()
            elif np.linalg.det(self.m) == 0:
                self.type = "degenerate"
                self.axis, self.angle = None, None

    def parse_screw_symmetry(self, tol=1e-2):
        """
        If the point group symmetry is rotation, parse the screw vector

        Returns:
            0, 2_1, 3_1, 3_2, 4_1, 4_2, 4_3, 6_1, 6_2, 6_3, 6_4, 6_5
        """
        # only count the translation on the given axis
        vec = self.translation_vector.copy()
        if np.isclose(abs(np.dot(self.axis, np.array([1, 0, 0]))), 1):
            vec[1] = 0
            vec[2] = 0
        elif np.isclose(abs(np.dot(self.axis, np.array([0, 1, 0]))), 1):
            vec[0] = 0
            vec[2] = 0
        elif np.isclose(abs(np.dot(self.axis, np.array([0, 0, 1]))), 1):
            vec[0] = 0
            vec[1] = 0
        else:
            vec = 0
        # No screw symmetry for other directios for tetragonal????

        if np.linalg.norm(vec) < tol:
            return str(self.order)  # 2, 3, 4, 6
        else:
            trans = np.sum(vec)
            if self.order == 2:
                return "2_1"
            elif self.order == 3:
                if abs(self.angle / trans - 2 * np.pi) < tol:
                    return "3_1"
                else:
                    return "3_2"
            elif self.order == 4:
                if abs(trans) < tol:  #
                    return "4_2"
                elif abs(self.angle / trans - 2 * np.pi) < tol:
                    return "4_1"
                elif abs(self.angle / trans - np.pi) < tol or abs(self.angle / trans - 3 * np.pi) < tol:
                    return "4_2"
                elif abs(self.angle / trans - 2 / 3 * np.pi) < tol or abs(self.angle / trans - 6 * np.pi) < tol:
                    return "4_3"
            elif self.order == 6:
                if abs(trans) < tol:  #
                    return "6_3"
                elif abs(self.angle / trans - 2 * np.pi) < tol:
                    return "6_1"
                elif abs(self.angle / trans - np.pi) < tol or abs(self.angle / trans - 5 / 2 * np.pi) < tol:
                    return "6_2"
                elif abs(self.angle / trans - 2 / 3 * np.pi) < tol or abs(self.angle / trans - 10 / 3 * np.pi) < tol:
                    return "6_3"
                elif abs(self.angle / trans - 1 / 2 * np.pi) < tol or abs(self.angle / trans - 5 * np.pi) < tol:
                    return "6_4"
                elif abs(self.angle / trans - 2 / 5 * np.pi) < tol or abs(self.angle / trans - 10 * np.pi) < tol:
                    return "6_5"

        print("Cannot assign symbol", self.angle, trans)
        return None

    def parse_glide_symmetry(self, tol=1e-2):
        """
        If the point group symmetry is rotation, parse the screw vector
        Returns:
            m, a, b, c, n, d
        """
        if self.rotation_order > 2:
            return "-" + str(self.rotation_order)
        elif abs(self.angle - np.pi) > tol:
            return "m"  # just indicate
        else:
            vec = self.translation_vector.copy()
            if np.isclose(abs(np.dot(self.axis, np.array([1, 0, 0]))), 1):
                vec[0] = 0
            elif np.isclose(abs(np.dot(self.axis, np.array([0, 1, 0]))), 1):
                vec[1] = 0
            elif np.isclose(abs(np.dot(self.axis, np.array([0, 0, 1]))), 1):
                vec[2] = 0

            if np.linalg.norm(vec) < tol:
                return "m"
            else:
                if np.linalg.norm(vec - np.array([1 / 2, 0, 0])) < tol:
                    return "a"
                elif np.linalg.norm(vec - np.array([0, 1 / 2, 0])) < tol:
                    return "b"
                elif np.linalg.norm(vec - np.array([0, 0, 1 / 2])) < tol:
                    return "c"
                elif (
                    np.linalg.norm(vec - np.array([0, 1 / 2, 1 / 2])) < tol
                    or np.linalg.norm(vec - np.array([1 / 2, 0, 1 / 2])) < tol
                    or np.linalg.norm(vec - np.array([1 / 2, 1 / 2, 0])) < tol
                ):
                    return "n"
                elif np.linalg.norm(vec - np.array([1 / 2, 1 / 2, 1 / 2])) < tol:
                    if np.isclose(abs(np.dot(self.axis, np.array([0, -0.7071, 0.7071]))), 1) or np.isclose(
                        abs(np.dot(self.axis, np.array([-0.7071, 0, 0.7071]))), 1
                    ):
                        return "n"
                    else:
                        return "c"
                else:
                    return "d"

    def parse_axis(self):
        """
        parse if the axis follows the standard convention
        """
        ax = self.axis
        ax /= np.linalg.norm(ax)
        for direction in all_sym_directions:
            normed = direction / np.linalg.norm(direction)
            # print(direction, np.dot(direction, ax))
            if np.isclose(np.dot(normed, ax), 1):
                return True
            elif np.isclose(np.dot(normed, ax), -1):
                return False
        raise ValueError("Cannot find the symmetry direction", ax)

    def __str__(self):
        """
        A custom printing string for the object. The type, order, angle, and
        axis are printed. Converts values close to 0 to 0 for readability. Also
        only prints the real part of the axis.
        """
        # Avoid printing '-0.' instead of '0.'
        if self.axis is not None and len(self.axis) == 3:
            for i, x in enumerate(self.axis):
                if np.isclose(x, 0):
                    self.axis[i] = 0.0
        return (
            "~~ Operation: "
            + self.op.as_xyz_str()
            + " ~~"
            + "\nType: "
            + str(self.type)
            + "\nOrder: "
            + str(self.order)
            + "\nAngle: "
            + str(self.angle)
            + "\nAxis: "
            + str(np.real(self.axis))
        )

    def is_conjugate(self, op2):
        """
        Returns whether or not another operation is conjugate (the same
        operation in a different reference frame). Rotations with the same order
        will not always return True. For example, a 5/12 and 1/12 rotation will
        not be considered conjugate.

        Args:
            op2: a SymmOp or OperationAnalyzer object to compare with

        Returns:
            True if op2 is conjugate to self.op, and False otherwise
        """
        if type(op2) != OperationAnalyzer:
            opa2 = OperationAnalyzer(op2)
            if opa2.type == self.type:
                if self.type == "rotation" or self.type == "rotoinversion":
                    ratio = self.angle / opa2.angle
                    if np.isclose(np.fabs(ratio), 1.0, atol=1e-2):
                        return True
                    return None
                elif self.type == "identity" or self.type == "inversion":
                    return True
                return None
            else:
                return False
        else:
            if op2.type == self.type:
                if self.type == "rotation" or self.type == "rotoinversion":
                    ratio = self.angle / op2.angle
                    if np.isclose(ratio, 1.0, atol=1e-2):
                        return True
                    return None
                elif self.type == "identity" or self.type == "inversion":
                    return True
                return None
            else:
                return False

    def are_conjugate(op1, op2):
        """
        Returns whether or not two operations are conjugate (the same
        operation in a different reference frame). Rotations with the same order
        will not always return True. For example, a 5/12 and 1/12 rotation will
        not be considered conjugate.

        Args:
            op1: a SymmOp or OperationAnalyzer object
            op2: a SymmOp or OperationAnalyzer object to compare with op1

        Returns:
            True if op2 is conjugate to op1, and False otherwise
        """
        if type(op1) != OperationAnalyzer:
            opa1 = OperationAnalyzer(op1)
        return opa1.is_conjugate(op2)


def find_ids(coords, ref, tol=1e-3):
    """
    find the refernce ids that can match
    """
    ids = []
    # print('ref', ref)
    for coord in coords:
        diffs = ref - coord
        diffs -= np.rint(diffs)
        norms = np.linalg.norm(diffs, axis=1)
        # print(norms, diffs)
        for i, norm in enumerate(norms):
            if norm < tol and i not in ids:
                ids.append(i)
                break
    return ids


def get_best_match(positions, ref, cell):
    """
    find the best match with the reference from a set of positions

    Args:
        positions: N*3 array
        ref: 1*3 array
        cell: cell matrix 3*3 array

    Returns:
        position: matched position
        id: matched id
    """
    diffs = positions - ref
    diffs -= np.rint(diffs)
    diffs = np.dot(diffs, cell)
    dists = np.linalg.norm(diffs, axis=1)
    id = np.argmin(dists)
    return positions[id], dists[id]


# Test Functionality
if __name__ == "__main__":
    # ----------------------------------------------------
    from pymatgen.symmetry.analyzer import generate_full_symmops

    rot = Rotation.from_rotvec(np.pi / 6 * np.array([1, 0, 0]))
    op = SymmOp.from_rotation_and_translation(rot.as_matrix(), np.array([1, 0, 0]))
    ops = [op]

    symm_m = generate_full_symmops(ops, 1e-3)
    for op in symm_m:
        opa = OperationAnalyzer(op)
        print(opa.order)



"""
DFS sampler
"""

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING

import numpy as np
from numpy.random import Generator
from pymatgen.analysis.structure_matcher import StructureMatcher

from pyxtal.optimize.base import GlobalOptimize

if TYPE_CHECKING:
    from pyxtal.lattice import Lattice
    from pyxtal.molecule import pyxtal_molecule


class DFS(GlobalOptimize):
    """
    Standard Population algorithm

    Args:
        smiles (str): smiles string
        workdir (str): path of working directory
        sg (int or list): space group number or list of spg numbers
        tag (string): job prefix
        ff_opt (bool): activate on the fly FF mode
        ff_style (str): automated force style (`gaff` or `openff`)
        ff_parameters (str or list): ff parameter xml file or list
        reference_file (str): path of reference xml data for FF training
        N_gen (int): number of generation (default: `10`)
        N_pop (int): number of populations (default: `10`)
        N_survival (int): number of survivals (default: `20`)
        N_cpu (int): number of cpus for parallel calculation (default: `1`)
        cif (str): cif file name to store all structure information
        block: block mode
        num_block: list of blocks
        compositions: list of composition, (default is [1]*Num_mol)
        lattice (bool): whether or not supply the lattice
        torsions: list of torsion angle
        molecules (list): list of pyxtal_molecule objects
        sites (list): list of wp sites, e.g., [['4a']]
        use_hall (bool): whether or not use hall number (default: False)
        skip_ani (bool): whether or not use ani or not (default: True)
        eng_cutoff (float): the cutoff energy for FF training
        E_max (float): maximum energy defined as an invalid structure
        verbose (bool): show more details
        use_mpi (bool): if use mpi
    """

    def __init__(
        self,
        smiles: str,
        workdir: str,
        sg: int | list,
        tag: str = 'test',
        info: dict[any, any] | None = None,
        ff_opt: bool = False,
        ff_style: str = "openff",
        ff_parameters: str = "parameters.xml",
        reference_file: str = "references.xml",
        ref_criteria: dict[any, any] | None = None,
        N_gen: int = 10,
        N_pop: int = 10,
        N_cpu: int = 1,
        N_survival: int = 20,
        cif: str | None = None,
        block: list[any] | None = None,
        num_block: list[any] | None = None,
        composition: list[any] | None = None,
        lattice: Lattice | None = None,
        torsions: list[any] | None = None,
        molecules: list[pyxtal_molecule] | None = None,
        sites: list[any] | None = None,
        use_hall: bool = False,
        skip_ani: bool = True,
        factor: float = 1.1,
        eng_cutoff: float = 5.0,
        E_max: float = 1e10,
        verbose: bool = False,
        random_state: int | None = None,
        max_time: float | None = None,
        matcher: StructureMatcher | None = None,
        early_quit: bool = False,
        check_stable: bool = False,
        use_mpi: bool = False,
        pre_opt: bool = False,
    ):
        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        # POPULATION parameters:
        self.N_gen = N_gen
        self.N_pop = N_pop
        self.N_survival = N_survival
        self.verbose = verbose
        self.name = 'DFS'

        # initialize other base parameters
        GlobalOptimize.__init__(
            self,
            smiles,
            workdir,
            sg,
            tag,
            info,
            ff_opt,
            ff_style,
            ff_parameters,
            reference_file,
            ref_criteria,
            N_cpu,
            cif,
            block,
            num_block,
            composition,
            lattice,
            torsions,
            molecules,
            sites,
            use_hall,
            skip_ani,
            factor,
            eng_cutoff,
            E_max,
            random_state,
            max_time,
            matcher,
            early_quit,
            check_stable,
            use_mpi,
            pre_opt,
        )

        # Setup the stats [N_gen, Npop, (E, matches)]
        self.stats = np.zeros([self.N_gen, self.N_pop, 2])
        self.stats[:, :, 0] = self.E_max

        if self.rank == 0:
            strs = self.full_str()
            self.logging.info(strs)
            print(strs)

    def full_str(self):
        s = str(self)
        s += "\nMethod    : Stochastic Depth First Sampling"
        s += f"\nGeneration: {self.N_gen:4d}"
        s += f"\nPopulation: {self.N_pop:4d}"
        # The rest base information from now on
        return s

    def _run(self, pool=None):
        """
        The main code to run DFS prediction

        Returns:
            success_rate or None
        """
        # Related to the FF optimization
        N_added = 0
        success_rate = 0
        cur_survivals = [0] * self.N_pop  # track the survivals
        hist_best_xtals = [None] * self.N_pop
        hist_best_engs = [self.E_max] * self.N_pop
        print(f"Rank {self.rank} starts DFS in {self.tag}")

        for gen in range(self.N_gen):
            self.generation = gen
            cur_xtals = None
            self.logging.info(f"Gen {gen} starts in Rank {self.rank}")

            if self.rank == 0:
                print(f"\nGeneration {gen:d} starts")
                self.logging.info(f"Generation {gen:d} starts")
                t0 = time()
                # Initialize structure and tags
                cur_xtals = [(None, "Random")] * self.N_pop

                # DFS update
                if gen > 0:
                    count = 0
                    mid_E = np.median(engs)

                    for id in range(self.N_pop):

                        # select the structures for further mutation
                        min_E = min([engs[id], hist_best_engs[id]])
                        if min_E < mid_E and cur_survivals[id] < self.N_survival:

                            if self.random_state.random() < 0.7:
                                source = prev_xtals[id][0]
                            else:
                                source = hist_best_xtals[id]

                            if source is not None:
                                cur_xtals[id] = (source, "Mutation")
                                cur_survivals[id] += 1
                                # Forget about the local best
                                if cur_survivals[id] == self.N_survival:
                                    hist_best_engs[id] = engs[id]
                                    hist_best_xtals[id] = prev_xtals[id][0]
                                count += 1

                        # Reset it to 0
                        if cur_xtals[id][1] == "Random":
                            cur_survivals[id] = 0

            # broadcast
            if self.use_mpi:
                cur_xtals = self.comm.bcast(cur_xtals, root=0)

            # Local optimization
            gen_results = self.local_optimization(cur_xtals, pool=pool)
            self.logging.info(f"Rank {self.rank} finishes local_opt.")

            prev_xtals = None
            if self.rank == 0:
                # pass results, summary_and_ranking
                cur_xtals, matches, engs = self.gen_summary(t0,
                                        gen_results, cur_xtals)

                # update hist_best
                for id, (xtal, _) in enumerate(cur_xtals):
                    if xtal is not None:
                        eng = xtal.energy / sum(xtal.numMols)
                        if eng < hist_best_engs[id]:
                            hist_best_engs[id] = eng
                            hist_best_xtals[id] = xtal

                # Save the reps for next move
                prev_xtals = cur_xtals  # ; print(self.engs)

            # broadcast
            if self.use_mpi:
                prev_xtals = self.comm.bcast(prev_xtals, root=0)
                self.logging.info(f"Gen {gen} bcast in Rank {self.rank}")

            # Update the FF parameters if necessary
            if self.ff_opt:
                N_added = self.update_ff_paramters(cur_xtals, engs, N_added)
            else:
                quit = False
                if self.rank == 0:
                    if self.ref_pmg is not None:
                        success_rate = self.success_count(cur_xtals, matches)
                        if self.early_termination(success_rate):
                            quit = True

                    elif self.ref_pxrd is not None:
                        self.count_pxrd_match(cur_xtals, matches)

                # quit the loop
                if self.use_mpi:
                    quit = self.comm.bcast(quit, root=0)
                    self.comm.Barrier()
                    self.logging.info(f"Gen {gen} Finish in Rank {self.rank}")

                # Ensure that all ranks exit
                if quit:
                    self.logging.info(f"Early Termination in Rank {self.rank}")
                    return success_rate

        return success_rate


if __name__ == "__main__":
    import argparse
    import os

    from pyxtal.db import database

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-g",
        "--gen",
        dest="gen",
        type=int,
        default=10,
        help="Number of generation, optional",
    )
    parser.add_argument(
        "-p",
        "--pop",
        dest="pop",
        type=int,
        default=10,
        help="Population size, optional",
    )
    parser.add_argument("-n", "--ncpu", dest="ncpu", type=int,
                        default=1, help="cpu number, optional")
    parser.add_argument("--ffopt", action="store_true",
                        help="enable ff optimization")

    options = parser.parse_args()
    gen = options.gen
    pop = options.pop
    ncpu = options.ncpu
    ffopt = options.ffopt
    db_name, name = "pyxtal/database/test.db", "ACSALA"
    wdir = name
    os.makedirs(wdir, exist_ok=True)
    os.makedirs(wdir + "/calc", exist_ok=True)

    db = database(db_name)
    row = db.get_row(name)
    xtal = db.get_pyxtal(name)
    smile, wt, spg = row.mol_smi, row.mol_weight, row.space_group.replace(
        " ", "")
    chm_info = None
    if not ffopt:
        if "charmm_info" in row.data:
            # prepare charmm input
            chm_info = row.data["charmm_info"]
            with open(wdir + "/calc/pyxtal.prm", "w") as prm:
                prm.write(chm_info["prm"])
            with open(wdir + "/calc/pyxtal.rtf", "w") as rtf:
                rtf.write(chm_info["rtf"])
        else:
            # Make sure we generate the initial guess from ambertools
            if os.path.exists("parameters.xml"):
                os.remove("parameters.xml")
    # load reference xtal
    pmg0 = xtal.to_pymatgen()
    if xtal.has_special_site():
        xtal = xtal.to_subgroup()
    N_torsion = xtal.get_num_torsions()

    # GO run
    t0 = time()
    go = DFS(
        smile,
        wdir,
        xtal.group.number,
        name.lower(),
        info=chm_info,
        ff_style="openff",  # 'gaff',
        ff_opt=ffopt,
        N_gen=gen,
        N_pop=pop,
        N_cpu=ncpu,
        cif="pyxtal.cif",
    )

    suc_rate = go.run(pmg0)
    print(f"CSD {name:s} in Gen {go.generation:d}")

    if len(go.matches) > 0:
        best_rank = go.print_matches()
        mytag = f"True {best_rank:d}/{go.N_struc:d} Succ_rate: {suc_rate:7.4f}%"
    else:
        mytag = f"False 0/{go.N_struc:d}"

    eng = go.min_energy
    t1 = int((time() - t0)/60)
    strs = "Final {:8s} [{:2d}]{:10s} ".format(name, sum(xtal.numMols), spg)
    strs += "{:3d}m {:2d} {:6.1f}".format(t1, N_torsion, wt)
    strs += "{:12.3f} {:20s} {:s}".format(eng, mytag, smile)
    print(strs)



import numpy as np
from random import choice
from ase.units import kB
from copy import deepcopy
import pymatgen.analysis.structure_matcher as sm
from pyxtal import pyxtal
from pyxtal.representation import representation
from pyxtal.molecule import find_id_from_smile
from htocsp.util import unique_rows

class BasinHopping():
    """
    Basin hopping algorithm.
    Wales and Doye, J. Phys. Chem. A, vol 101 (1997) 5111-5116
    """

    def __init__(self, x, sg, smile, atom_info, workdir, tag,
                 optimizer, kT=0.1*kB, dr=0.1,
                 composition=None, lattice=None, torsions=None,
                ):
        """
        Args:
            x: 1D list [a, b, c, ....]
            sg: space group number
            smile:
            atom_info: a dictinonary for atom properties (e.g., label, charge)
            workdir:
            sg:
            optimizer:
            kT: boltzmann factor (default: `0.1kB`)
            dr: magnitude of mutation (default: `0.2`)
        """
        # BH parameters
        self.kT = kT
        self.dr = dr

        # setup for local optimization
        self.lattice = lattice
        if lattice is None:
            self.opt_lat = True
        else:
            self.opt_lat = False
        self.optimizer = optimizer
        self.smile = smile
        self.atom_info = atom_info
        self.workdir = workdir
        self.tag = tag
        self.sg = sg
        self.torsions = torsions
        self.smiles = self.smile.split('.') #list
        if composition is None:
            self.composition = [1] * len(self.smiles)
        else:
            self.composition = composition
        self.N_torsion = 0
        for smi, comp in zip(self.smiles, self.composition):
            self.N_torsion += len(find_id_from_smile(smi))*comp
        print(self)

        # initial guess
        self.x = x

        # all structures
        self.reps = []

    def __str__(self):
        s = "\n------Basin Hopping Algorithm------"
        s += "\nkT      : {:6.2f}".format(self.kT)
        s += "\ndr      : {:6.2f}".format(self.dr)
        s += "\nsmile   : {:s}".format(self.smile)
        s += "\ndiretory: {:s}".format(self.workdir)
        s += "\nZprime    : {:s}".format(str(self.composition))
        s += "\nN_torsion : {:d}".format(self.N_torsion)
        s += "\nsg        : {:s}".format(str(self.sg))
        s += "\nopt_lat   : {:s}\n".format(str(self.opt_lat))
        return s

    def __repr__(self):
        return str(self)

    def run(self, steps=100, show=False):
        """
        Hop the basins for defined number of steps.
        """
        # Initial solution
        self.x, Eo, strs = self.get_energy(self.x)
        strs = "{:3d}".format(0) + strs
        print(strs)

        #print("FIRST", self.x[:5])
        for step in range(steps):
            En = None
            while En is None:
                rn = self.move()
                rn, En, strs = self.get_energy(rn)
            #print(Eo, En, np.exp((Eo-En)/self.kT))

            accept = False
            diff = Eo-En
            if diff>0:
                if abs(diff) > 5e-3: #identical
                    accept = True
            else:
                if np.exp(diff/self.kT)>np.random.uniform():
                    accept = True
            if accept:
                self.x = deepcopy(rn)
                self.reps.append(rn)
                Eo = En
                strs += "+++"
            strs = "{:3d}".format(step+1) + strs
            if show and accept: print(strs)

        print("Calculation is complete")

    def move(self):
        # displace coordinates
        x = self.x
        dr = self.dr
        # perturb cell
        x_new = deepcopy(x)
        if self.opt_lat:
            sg = x[0][0]
            disp_cell = np.random.uniform(-1., 1., len(x[0])-2)
            x_new[0][2:] *= (1+dr*disp_cell)

            #flip the inclination angle
            if 3 <= sg <= 15 and np.random.random() > 0.7:
                if abs(90-x_new[0][-1])<15:
                    x_new[0][-1] = 180-x_new[0][-1]

        # perturb molecules
        for i in range(1, len(x)):
            disp_mol = np.random.uniform(-1., 1., len(x_new[i])-1)
            x_new[i][:-1] *= (1+dr*disp_mol)
            # change the orientation and torsions
            for j in range(3, len(x_new[i])-1):
                rad_num = np.random.random()
                if rad_num < 0.25:
                    x_new[i][j] += choice([45.0, 90.0])
                elif rad_num < 0.5:
                    x_new[i][j] *= -1
        return x_new

    def get_energy(self, x):
        """
        Return the energy of the nearest local minimum.
        """
        res = self.optimizer(x, self.smiles, self.composition, self.atom_info,\
                             self.workdir, self.tag, opt_lat=self.opt_lat)
        #print(res); import sys; sys.exit()
        if res is not None:
            rep, eng = res['rep'], res['energy']
            myrep = representation(rep, self.smiles)
            strs = myrep.to_string(None, eng)
            return rep, eng, strs
        return None, None, None


    def check_ref(self, xtal):
        """
        check if ground state structure is found

        Args:
            - xtal: pyxtal object
        """

        print("check if ground state structure is found")
        rep = representation.from_pyxtal(xtal)
        print("\n" + rep.to_string() + ' reference')
        pmg0 = xtal.to_pymatgen()
        pmg0.remove_species('H')

        for d in self.reps:
            rep0 = representation(d, self.smiles)
            s = rep0.to_pyxtal(composition=self.composition)
            pmg_struc = s.to_pymatgen()
            pmg_struc.remove_species('H')
            found = sm.StructureMatcher().fit(pmg_struc, pmg0)
            if found:
                strs += " " + str(found)
                print(strs)
                break

def basin_hopping_runner(smile, atom_info, workdir, tag, sg, comp, optimize, \
                         kT, rep=None, steps=100, xtal_ref=None,
                         show=False, seed=None):
    smiles = smile.split('.')
    if rep is None:
        g = Group(sg)[0]
        mult = len(g)
        letter = g.letter
        numIons = [c*mult for c in comp]
        sites = []
        for c in comp:
            sites.append([str(mult)+letter]*c)

        while True:
            if seed is None:
                np.random.seed()
            diag = choice([True, False])
            xtal = pyxtal(molecular=True)
            xtal.from_random(3, sg, smiles, numIons, force_pass=True, diag=diag)
            rep = representation.from_pyxtal(xtal)
            print(rep.x, smiles, composition, workdir, tag)
            print(atom_info)
            res = optimize(rep.x, smiles, comp, atom_info, workdir, tag)
            if res is not None:
                break
    elif isinstance(rep, str):
        rep = representation.from_string(rep, smiles, comp)

    go = BasinHopping(rep.x, sg, smile, atom_info, workdir, tag, optimize, kT, composition=comp)
    print('  0' + rep.to_string())
    go.run(steps, show)

    if xtal_ref is not None:
        go.check_ref(xtal_ref)


if __name__ == "__main__":
    from pyxtal import pyxtal
    from htocsp.localopt import optimize
    from htocsp.util import parse_mol
    from htocsp.interfaces.charmm import RTF, PRM
    from optparse import OptionParser
    import random

    parser = OptionParser()
    parser.add_option("-t", "--tag", dest="tag", default='job_bh',
                      help="job tag, optional")
    parser.add_option("-k", "--kT", dest="kT", type=float, default=0.5,
                      help="kT factor, optional")
    parser.add_option("-s", "--step", dest="step", type=int, default=10,
                      help="number of steps, optional")
    parser.add_option("-n", "--npar", dest="npar", type=int, default=1,
                      help="number of parallel computation, optional")
    parser.add_option("-r", "--repeats", dest="repeats", type=int, default=1,
                      help="repeat")


    (options, args) = parser.parse_args()
    random.seed(10)
    seed = 4 #needed for pyxtal orientation
    np.random.seed(seed)
    smile, workdir, name, sg, f, composition = 'CC(=O)OC1=CC=CC=C1C(=O)O', 'BH', 'ACSALA', 14, 'benchmarks/aspirin.cif', [1]
    xtal_ref = pyxtal(molecular=True)
    xtal_ref.from_seed(f, molecules=[smile+'.smi'])

    #print(xtal_ref)
    p_mol, rtf, prm = parse_mol(name=name, smi=smile)
    charmm_info = {}
    charmm_info['label']   = [p_mol.site_properties['charmm_label']]
    charmm_info['resName'] = [p_mol.site_properties['resName'][0]]
    prm = PRM(prm.split('\n')); rtf = RTF(rtf.split('\n'))
    charmm_info['prm'] = prm.to_string(); charmm_info['rtf'] = rtf.to_string()
    prm = open(workdir+'/pyxtal.prm', 'w'); prm.write(charmm_info['prm']); prm.close()
    rtf = open(workdir+'/pyxtal.rtf', 'w'); rtf.write(charmm_info['rtf']); rtf.close()

    rep = "14 0 11.43  6.49 11.19 83.31 1  0.77  0.57  0.53 48.55 24.31 145.9 -77.85 -4.40 170.9 0" #None
    tag = options.tag
    kT = options.kT
    steps = options.step
    npar = options.npar
    repeats = options.repeats

    basin_hopping_runner(smile, charmm_info, workdir, tag, sg, composition,\
                         optimize, kT, rep, steps, xtal_ref,
                         show=True, seed=seed)



from __future__ import division
import numpy as np
from scipy.special import sph_harm, spherical_in
from ase import Atoms
from ase.neighborlist import NeighborList


class SO3:
    '''
    A class to generate the SO3 power spectrum components
    based off of the Gaussian atomic neighbor density function
    defined in "On Representing Atomic Environments".

    args:
        nmax: int, degree of radial expansion
        lmax: int, degree of spherical harmonic expansion
        rcut: float, cutoff radius for neighbor calculation
        alpha: float, gaussian width parameter
        weight_on: bool, if True, the neighbors with different type will be counted as negative
    '''

    def __init__(self, nmax=3, lmax=3, rcut=3.5, alpha=2.0,
                 weight_on=False):
        # populate attributes
        self.nmax = nmax
        self.lmax = lmax
        self.rcut = rcut
        self.alpha = alpha
        self._type = "SO3"
        self.cutoff_function = 'cosine'
        self.weight_on = weight_on
        self.ncoefs = self.nmax*(self.nmax+1)//2*(self.lmax+1)
        self.tril_indices = np.tril_indices(self.nmax, k=0)
        self.ls = np.arange(self.lmax+1)
        self.norm = np.sqrt(2*np.sqrt(2)*np.pi/np.sqrt(2*self.ls+1))
        self.keys = ['keys', '_nmax', '_lmax', '_rcut', '_alpha',
                     '_cutoff_function', 'weight_on',
                     'ncoefs', 'ls', 'norm', 'tril_indices', 'args']
        self.args = (self.nmax, self.lmax, self.rcut, self.alpha, self._cutoff_function)

    def __str__(self):
        s = "SO3 descriptor with Cutoff: {:6.3f}".format(self.rcut)
        s += " lmax: {:d}, nmax: {:d}, alpha: {:.3f}\n".format(self.lmax, self.nmax, self.alpha)
        return s

    def __repr__(self):
        return str(self)

    def load_from_dict(self, dict0):
        self.nmax = dict0["nmax"]
        self.lmax = dict0["lmax"]
        self.rcut = dict0["rcut"]
        self.alpha = dict0["alpha"]
        self.derivative = dict0["derivative"]

    def save_dict(self):
        """
        save the model as a dictionary in json
        """
        dict = {"nmax": self.nmax,
                "lmax": self.lmax,
                "rcut": self.rcut,
                "alpha": self.alpha,
                "derivative": self.derivative,
                "_type": "SO3",
               }
        return dict

    @property
    def nmax(self):
        return self._nmax

    @nmax.setter
    def nmax(self, nmax):
        if isinstance(nmax, int) is True:
            if nmax < 1:
                raise ValueError('nmax must be greater than or equal to 1')
            if nmax > 11:
                raise ValueError('nmax > 11 yields complex eigenvalues')
            self._nmax = nmax
        else:
            raise ValueError('nmax must be an integer')

    @property
    def lmax(self):
        return self._lmax

    @lmax.setter
    def lmax(self, lmax):
        if isinstance(lmax, int) is True:
            if lmax < 0:
                raise ValueError('lmax must be greater than or equal to zero')
            elif lmax > 32:
                raise NotImplementedError('support a maxmimum l=32 for spherical harmonics')
            self._lmax = lmax
        else:
            raise ValueError('lmax must be an integer')

    @property
    def rcut(self):
        return self._rcut

    @rcut.setter
    def rcut(self, rcut):
        if isinstance(rcut, float) or isinstance(rcut, int):
            if rcut <= 0:
                raise ValueError('rcut must be greater than zero')
            self._rcut = rcut
        else:
            raise ValueError('rcut must be a float')

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        if isinstance(alpha, float) or isinstance(alpha, int):
            if alpha <= 0:
                raise ValueError('alpha must be greater than zero')
            self._alpha = alpha
        else:
            raise ValueError('alpha must be a float')

    @property
    def cutoff_function(self):
        return self._cutoff_function

    @cutoff_function.setter
    def cutoff_function(self, cutoff_function):
        self._cutoff_function = Cosine

    def clear_memory(self):
        '''
        Clears all non essential attributes for the calculator
        '''
        attrs = list(vars(self).keys())
        for attr in attrs:
            if attr not in self.keys:
                delattr(self, attr)
        return

    def init_atoms(self, atoms, atom_ids=None):
        """
        initilize atoms related attributes
        """
        self._atoms = atoms
        self.natoms = len(atoms)
        self.build_neighbor_list(atom_ids)

    def compute_p(self, atoms, atom_ids=None, return_CN=False):
        """
        Compute the powerspectrum function

        Args:
            atoms: ase atoms object
            atom_ids: optional list of atomic indices

        Returns:
            p array (N, M)
        """

        if atom_ids is None: atom_ids = range(len(atoms))
        self.init_atoms(atoms, atom_ids)
        plist = np.zeros((len(atom_ids), self.ncoefs), dtype=np.float64)
        dist = np.zeros((len(atom_ids), 2))

        if len(self.neighborlist) > 0:
            cs = compute_cs(self.neighborlist, *self.args)
            cs *= self.atomic_weights[:, np.newaxis, np.newaxis, np.newaxis]
            cs = np.einsum('inlm,l->inlm', cs, self.norm)

            # Get r_ij and compute C*np.conj(C)
            for _i, i in enumerate(atom_ids):
                centers = self.neighbor_indices[:, 0] == i
                CN = len(self.neighborlist[centers])
                if CN > 0:
                    ctot = cs[centers].sum(axis=0)
                    P = np.einsum('ijk,ljk->ilj', ctot, np.conj(ctot)).real
                    plist[_i] = P[self.tril_indices].flatten()
                    dist[_i, 0] += np.linalg.norm(self.neighborlist[centers], axis=1).max()
                    dist[_i, 1] += CN
        if return_CN:
            return plist, dist
        else:
            return plist


    def compute_dpdr(self, atoms, atom_ids=None):
        """
        Compute the powerspectrum function

        Args:
            atoms: ase atoms object
            atom_ids: optional list of atomic indices

        Returns:
            dpdr array (N, N, M, 3) and p array (N, M)
        """

        if atom_ids is None: atom_ids = range(len(atoms))
        self.init_atoms(atoms, atom_ids)
        p_list = np.zeros((len(atom_ids), self.ncoefs), dtype=np.float64)
        dp_list = np.zeros((len(atom_ids), self.natoms, self.ncoefs, 3), dtype=np.float64)

        if len(self.neighborlist) > 0:
            # get expansion coefficients and derivatives
            cs, dcs = compute_dcs(self.neighborlist, *self.args)

            # weight cs and dcs
            cs *= self.atomic_weights[:, np.newaxis, np.newaxis, np.newaxis]
            dcs *= self.atomic_weights[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
            cs = np.einsum('inlm,l->inlm', cs, self.norm)
            dcs = np.einsum('inlmj,l->inlmj', dcs, self.norm)
            #print('cs, dcs', self.neighbor_indices, cs.shape, dcs.shape)

            # cs: (N_ij, n, l, m)     => P (N_i, N_des)
            # dcs: (N_ij, n, l, m, 3) => dP (N_i, N_j, N_des, 3)
            # (n, l, m) needs to be merged to 1 dimension

            for _i, i in enumerate(atom_ids):
                # find atoms for which i is the center
                centers = self.neighbor_indices[:, 0] == i

                if len(self.neighbor_indices[centers]) > 0:
                    # total up the c array for the center atom
                    ctot = cs[centers].sum(axis=0) #(n, l, m)

                    # power spectrum P = c*c_conj
                    # eq_3 (n, n', l) eliminate m
                    P = np.einsum('ijk, ljk->ilj', ctot, np.conj(ctot)).real
                    p_list[_i] += P[self.tril_indices].flatten()

                    # gradient of P for each neighbor, eq_26
                    # (N_ijs, n, n', l, 3)
                    # dc * c_conj + c * dc_conj
                    dP = np.einsum('wijkn,ljk->wiljn', dcs[centers], np.conj(ctot))
                    dP += np.conj(np.transpose(dP, axes=[0, 2, 1, 3, 4]))
                    dP = dP.real

                    #print("shape of P/dP", P.shape, dP.shape)#; import sys; sys.exit()

                    # QZ: to check
                    ijs = self.neighbor_indices[centers]
                    for _id, j in enumerate(ijs[:, 1]):
                        tmp = dP[_id][self.tril_indices].flatten().reshape(self.ncoefs, 3)
                        dp_list[_i, j, :, :] += tmp
                        dp_list[_i, i, :, :] -= tmp

        return dp_list, p_list

    def compute_dpdr_5d(self, atoms, atom_ids=None):
        """
        Compute the powerspectrum function with respect to supercell

        Args:
            atoms: ase atoms object

        Returns:
            dpdr array (N, N, M, 3, 27) and p array (N, M)
        """
        if atom_ids is None: atom_ids = range(len(atoms))
        self.init_atoms(atoms, atom_ids)
        p_list = np.zeros((len(atom_ids), self.ncoefs), dtype=np.float64)
        dp_list = np.zeros((len(atom_ids), self.natoms, self.ncoefs, 3, 27), dtype=np.float64)

        if len(self.neighborlist) > 0:
            # get expansion coefficients and derivatives
            cs, dcs = compute_dcs(self.neighborlist, *self.args)

            # weight cs and dcs
            cs *= self.atomic_weights[:, np.newaxis, np.newaxis, np.newaxis]
            dcs *= self.atomic_weights[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
            cs = np.einsum('inlm,l->inlm', cs, self.norm)
            dcs = np.einsum('inlmj,l->inlmj', dcs, self.norm)
            #print('cs, dcs', self.neighbor_indices, cs.shape, dcs.shape)

            # cs: (N_ij, n, l, m)     => P (N_i, N_des)
            # dcs: (N_ij, n, l, m, 3) => dP (N_i, N_j, N_des, 3)
            # (n, l, m) needs to be merged to 1 dimension
            neigh_ids = np.arange(len(self.neighbor_indices))
            for _i, i in enumerate(atom_ids):
                # find atoms for which i is the center
                pair_ids = neigh_ids[self.neighbor_indices[:, 0] == i]
                if len(pair_ids) > 0:
                    ctot = cs[pair_ids].sum(axis=0) #(n, l, m)
                    dctot = dcs[pair_ids].sum(axis=0)
                    # power spectrum P = c * c_conj
                    P = np.einsum('ijk, ljk->ilj', ctot, np.conj(ctot)).real
                    p_list[_i] += P[self.tril_indices].flatten()

                    # loop over each pair
                    for pair_id in pair_ids:
                        (_, j, cell_id) = self.neighbor_indices[pair_id]
                        # map from (x, y, z) to (0, 27)

                        # (N_ijs, n, n', l, 3)
                        # dp = dc * c_conj + c * dc_conj
                        dP = np.einsum('ijkn, ljk->iljn', dcs[pair_id], np.conj(ctot))
                        dP += np.einsum('ijkn, ljk->iljn', np.conj(dcs[pair_id]), ctot)

                        dP = dP.real[self.tril_indices].flatten().reshape(self.ncoefs, 3)
                        #print(cs[pair_id].shape, dcs[pair_id].shape, dP.shape)

                        dp_list[_i, j, :, :, cell_id] += dP
                        dp_list[_i, i, :, :, 13] -= dP

        return dp_list, p_list


    def calculate(self, atoms, atom_ids=None, derivative=False):
        '''
        API for Calculating the SO(3) power spectrum components of the
        smoothened atomic neighbor density function

        Args:
            atoms: an ASE atoms object corresponding to the desired
                   atomic arrangement
            derivative: bool, whether to calculate the gradient of not
        '''
        p_list = None
        dp_list = None
        if derivative:
            dp_list, p_list = self.compute_dpdr(atoms, atom_ids)
        else:
            p_list = self.compute_p(atoms, atom_ids)

        x = {'x': p_list,
             'dxdr': dp_list,
             'elements': list(atoms.symbols)}
        self.clear_memory()
        return x


    def build_neighbor_list(self, atom_ids=None):
        '''
        Builds a neighborlist for the calculation of bispectrum components for
        a given ASE atoms object given in the calculate method.
        '''
        atoms = self._atoms
        cell_matrix = atoms.get_cell()
        VECTORS = np.array([[x1, y1, z1] for x1 in range(-1, 2) for y1 in range(-1, 2) for z1 in range(-1, 2)])
        neighbors = []
        neighbor_indices = []
        atomic_weights = []

        #if True: #atom_ids is None:
        if atom_ids is None:
            atom_ids = range(len(atoms))

            cutoffs = [self.rcut/2]*len(atoms)
            nl = NeighborList(cutoffs, self_interaction=False, bothways=True, skin=0.0)
            nl.update(atoms)

            for i in atom_ids:
                # get center atom position vector
                center_atom = atoms.positions[i]
                # get indices and cell offsets for each neighbor
                indices, offsets = nl.get_neighbors(i)
                #print(indices); import sys; sys.exit()

                for j, offset in zip(indices, offsets):
                    (x, y, z) = offset
                    cell_id = (x+1) * 9 + (y+1) * 3 + z + 1
                    pos = atoms.positions[j] + offset@cell_matrix - center_atom
                    if np.sum(np.abs(pos)) < 1e-3:
                        # to skip self
                        if cell_id == 13:
                            continue
                        # to prevent division by zero
                        else:
                            pos += 1e-3
                    neighbors.append(pos)
                    if self.weight_on and atoms[j].number != atoms[i].number:
                        factor = -1
                    else:
                        factor = 1
                    atomic_weights.append(factor*atoms[j].number)
                    neighbor_indices.append([i, j, cell_id])
        else:
            # A short cut version if we only compute the neighbors for a few atoms
            ref_pos = np.repeat(atoms.positions[:, :, np.newaxis], 27, axis=2)
            cell_shifts = np.dot(VECTORS, cell_matrix)
            ref_pos += cell_shifts.T[np.newaxis, :, :]
            ref_ids = np.arange(len(atoms))
            for i in atom_ids:
                center_atom = atoms.positions[i]

                # Compute distances relative to the center_atom across all cells
                dists = np.linalg.norm(ref_pos - center_atom[:, np.newaxis], axis=1)

                # Create mask for distances that are within the cutoff and greater than 1e-2
                mask = (dists > 1e-2) & (dists < self.rcut)

                # Use the mask to find neighbors in all cells at once
                valid_atoms, valid_cells = np.where(mask)
                # Append the neighbors and weights
                for j, cell_id in zip(valid_atoms, valid_cells):
                    neighbors.append(ref_pos[j, :, cell_id] - center_atom)
                    factor = -1 if self.weight_on and atoms[j].number != atoms[i].number else 1
                    atomic_weights.append(factor * atoms[j].number)
                    neighbor_indices.append([i, j, cell_id])
                #for cell_id in range(27):
                #    dists = np.linalg.norm(ref_pos[:, :, cell_id] - center_atom, axis=1)
                #    mask = (dists > 1e-2) & (dists < self.rcut)
                #    for j in ref_ids[mask]:
                #        neighbors.append(ref_pos[j, :, cell_id] - center_atom)
                #        factor = -1 if self.weight_on and atoms[j].number != atoms[i].number else 1
                #        atomic_weights.append(factor*atoms[j].number)
                #        neighbor_indices.append([i, j, cell_id])

        self.neighborlist = np.array(neighbors, dtype=np.float64)
        self.atomic_weights = np.array(atomic_weights, dtype=np.int64)
        self.neighbor_indices = np.array(neighbor_indices, dtype=np.int64)

def Cosine(Rij, Rc, derivative=False):
    # Rij is the norm
    if not derivative:
        result = 0.5 * (np.cos(np.pi * Rij / Rc) + 1.)
    else:
        result = -0.5 * np.pi / Rc * np.sin(np.pi * Rij / Rc)
    return result

def W(nmax):
    arr = np.zeros((nmax,nmax), np.float64)
    for alpha in range(1, nmax+1, 1):
        temp1 = (2*alpha+5)*(2*alpha+6)*(2*alpha+7)
        for beta in range(1, alpha+1, 1):
            temp2 = (2*beta+5)*(2*beta+6)*(2*beta+7)
            arr[alpha-1, beta-1] = np.sqrt(temp1*temp2)/(5+alpha+beta)/(6+alpha+beta)/(7+alpha+beta)
            arr[beta-1, alpha-1] = arr[alpha-1, beta-1]

    sinv = np.linalg.inv(arr)
    eigvals, V = np.linalg.eig(sinv)
    sqrtD = np.diag(np.sqrt(eigvals))
    arr[:,:] = np.dot(np.dot(V, sqrtD), np.linalg.inv(V))
    return arr

def phi(r, alpha, rcut):
    '''
    See g below
    '''
    return (rcut-r)**(alpha+2)/np.sqrt(2*rcut**(2*alpha+7)/(2*alpha+5)/(2*alpha+6)/(2*alpha+7))

def g(r, n, nmax, rcut, w):

    Sum = 0.0
    for alpha in range(1, nmax+1):
        Sum += w[n-1, alpha-1]*phi(r, alpha, rcut)

    return Sum

def GaussChebyshevQuadrature(nmax, lmax):
    NQuad = (nmax+lmax+1) * 10 #
    quad_array = np.zeros(NQuad, dtype=np.float64)
    for i in range(1, NQuad+1):
        # roots of Chebyshev polynomial of degree N
        x = np.cos((2*i-1)*np.pi/2/NQuad)
        quad_array[i-1] = x
    return quad_array, np.pi/NQuad

def compute_cs(pos, nmax, lmax, rcut, alpha, cutoff):
    """
    Compute expansion coefficients for a system based on the input positions.

    This function calculates the expansion coefficients for a set of atomic positions using
    Gauss-Chebyshev quadrature, spherical Bessel functions, and spherical harmonics. It is
    typically used in models that require high-dimensional projections of atomic environments.

    Args:
        pos (numpy.ndarray): An array of atomic positions (N x 3) where N is the number of atoms.
        nmax (int): Maximum radial quantum number used in the expansion.
        lmax (int): Maximum angular momentum quantum number used in the expansion.
        rcut (float): Cutoff radius for interactions and the radial expansion.
        alpha (float): Gaussian decay factor applied to the radial functions.
        cutoff (callable): A function to compute cutoff values for the radial distances.

    Returns:
        numpy.ndarray: A 4D array of expansion coefficients with shape (N_neighbors, nmax, lmax+1, 2*lmax+1),
        where `N_neighbors` is the number of neighbor atoms, `nmax` is the number of radial terms, and
        `lmax` and `m` correspond to angular momentum quantum numbers.
    """

    # 1. Init Overlap matrix, distances and quadrature points
    w = W(nmax)
    Ris = np.linalg.norm(pos, axis=1)  # (N_neighbors)
    GCQuadrature, weight = GaussChebyshevQuadrature(nmax, lmax)  # (Nquad)
    weight *= rcut / 2
    Quadrature = rcut / 2 * (GCQuadrature + 1)

    # 2. Rdial G functions
    Gs = np.zeros((nmax, len(Quadrature)), dtype=np.float64)  # (nmax, Nquad)
    for n in range(1, nmax + 1):
        Gs[n - 1, :] = g(Quadrature, n, nmax, rcut, w)

    Quad_Squared = Quadrature ** 2
    Gs *= Quad_Squared * np.exp(-alpha * Quad_Squared) * np.sqrt(1 - GCQuadrature**2) * weight

    # 3. Bessel functions
    BesselArgs = 2 * alpha * np.outer(Ris, Quadrature)  # (N_neighbors x Nquad)
    Bessels = np.zeros((len(Ris), len(Quadrature), lmax + 1), dtype=np.float64)  # (N_neighbors x Nquad x lmax+1)
    for l in range(lmax + 1):
        Bessels[:, :, l] = spherical_in(l, BesselArgs)

    # 4. Integration over radial coordinates using Einstein summation
    integral_array = np.einsum('ij,kjl->kil', Gs, Bessels)  # (N_neighbors x nmax x lmax+1)

    # 5. Gaussian factors for each atom and apply the cutoff function
    exparray = 4 * np.pi * np.exp(-alpha * Ris**2)  # (N_neighbors)
    cutoff_array = cutoff(Ris, rcut)
    np.multiply(exparray, cutoff_array, out=exparray)

    # 6. Compute the spherical harmonics for each atom
    # From Cartesian coordinates to spherical coordinates
    thetas = pos[:, 2] / Ris[:]
    thetas = np.arccos(thetas)
    phis = np.arctan2(pos[:, 1], pos[:, 0])
    msize = 2 * lmax + 1
    ylms = np.zeros((len(Ris), lmax + 1, msize), dtype=np.complex128)

    #for l in range(lmax + 1):
    #    for m in range(-l, l + 1):
    #        midx = msize // 2 + m
    #        ylms[:, l, midx] = sph_harm(m, l, phis, thetas)
    l_vals = np.repeat(np.arange(lmax + 1), [2 * l + 1 for l in range(lmax + 1)])
    m_vals = np.concatenate([np.arange(-l, l + 1) for l in range(lmax + 1)])
    midx_vals = msize // 2 + m_vals
    ylms[:, l_vals, midx_vals] = sph_harm(m_vals, l_vals, phis[:, np.newaxis], thetas[:, np.newaxis])


    # 7. Multiply Ylm and Gaussian factors
    Y_mul_innerprod = np.einsum('ijk,ilj->iljk', ylms, integral_array)
    C = np.einsum('i,ijkl->ijkl', exparray, Y_mul_innerprod)

    return C

def compute_dcs(pos, nmax, lmax, rcut, alpha, cutoff):
    """
    Compute exapnsion coefficients

    Args:
        pos:
        nmax (int):
        lmax (int):
        rcut (float):
        alpha (float):
        cutoff (callable):

    Returns:
        c(N_ij, nmax, lmax+1, 2lmax+1)
        dc(N_ij, nmax, lmax+1, 2lmax+1, 3) for each x,y,z
    """
    # compute the overlap matrix
    w = W(nmax)

    # get the norm of the position vectors
    Ris = np.linalg.norm(pos, axis=1) # (Nneighbors)

    # get unit vectors
    upos = pos/Ris[:,np.newaxis]

    # initialize Gauss Chebyshev Quadrature
    GCQuadrature, weight = GaussChebyshevQuadrature(nmax,lmax) #(Nquad)
    weight *= rcut/2
    # transform from (-1,1) to (0, rcut)
    Quadrature = rcut/2*(GCQuadrature+1)

    # compute the arguments for the bessel functions
    BesselArgs = 2*alpha*np.outer(Ris,Quadrature)#(Nneighbors x Nquad)

    # initalize the arrays for the bessel function values
    # and the G function values
    Bessels = np.zeros((len(Ris), len(Quadrature), lmax+1), dtype=np.float64) #(Nneighbors x Nquad x lmax+1)
    Gs = np.zeros((nmax, len(Quadrature)), dtype=np.float64) # (nmax, nquad)
    dBessels = np.zeros((len(Ris), len(Quadrature), lmax+1), dtype=np.float64) #(Nneighbors x Nquad x lmax+1)

    # compute the g values
    for n in range(1,nmax+1,1):
        Gs[n-1,:] = g(Quadrature, n, nmax, rcut,w)*weight

    # compute the bessel values
    for l in range(lmax+1):
        Bessels[:,:,l] = spherical_in(l, BesselArgs)
        dBessels[:,:,l] = spherical_in(l, BesselArgs, derivative=True)

    #(Nneighbors x Nquad x lmax+1) unit vector here
    gradBessels = np.einsum('ijk,in->ijkn',dBessels,upos)
    gradBessels *= 2*alpha

    # multiply with r for the integral
    gradBessels = np.einsum('ijkn,j->ijkn',gradBessels,Quadrature)

    # mutliply the terms in the integral separate from the Bessels
    Quad_Squared = Quadrature**2
    Gs *= Quad_Squared * np.exp(-alpha*Quad_Squared) * np.sqrt(1-GCQuadrature**2)

    # perform the integration with the Bessels
    integral_array = np.einsum('ij,kjl->kil', Gs, Bessels) # (Nneighbors x nmax x lmax+1)

    grad_integral_array = np.einsum('ij,kjlm->kilm', Gs, gradBessels)# (Nneighbors x nmax x lmax+1, 3)

    # compute the gaussian for each atom
    exparray = 4*np.pi*np.exp(-alpha*Ris**2) # (Nneighbors)

    gradexparray = (-2*alpha*Ris*exparray)[:,np.newaxis]*upos

    cutoff_array = cutoff(Ris, rcut)

    grad_cutoff_array = np.einsum('i,in->in',cutoff(Ris, rcut, True), upos)

    # get the spherical coordinates of each atom
    thetas = np.arccos(pos[:,2]/Ris[:])
    phis = np.arctan2(pos[:,1], pos[:,0])

    # the size changes temporarily for the derivative
    # determine the size of the m axis
    Msize = 2*(lmax+1)+1
    msize = 2*lmax + 1
    # initialize an array for the spherical harmonics and gradients
    #(Nneighbors, l, m, *3*)
    ylms = np.zeros((len(Ris), lmax+1+1, Msize), dtype=np.complex128)
    gradylms = np.zeros((len(Ris), lmax+1, msize, 3), dtype=np.complex128)
    # compute the spherical harmonics
    for l in range(lmax+1+1):
        for m in range(-l,l+1,1):
            midx = Msize//2 + m
            ylms[:,l,midx] = sph_harm(m, l, phis, thetas)


    for l in range(1, lmax+1):
        for m in range(-l, l+1, 1):
            midx = msize//2 + m
            Midx = Msize//2 + m
            # get gradient with recpect to spherical covariant components
            xcov0 = -np.sqrt(((l+1)**2-m**2)/(2*l+1)/(2*l+3))*l*ylms[:,l+1,Midx]/Ris

            if abs(m) <= l-1:
                xcov0 += np.sqrt((l**2-m**2)/(2*l-1)/(2*l+1))*(l+1)*ylms[:,l-1,Midx]/Ris


            xcovpl1 = -np.sqrt((l+m+1)*(l+m+2)/2/(2*l+1)/(2*l+3))*l*ylms[:,l+1,Midx+1]/Ris

            if abs(m+1) <= l-1:
                xcovpl1 -= np.sqrt((l-m-1)*(l-m)/2/(2*l-1)/(2*l+1))*(l+1)*ylms[:,l-1,Midx+1]/Ris


            xcovm1 = -np.sqrt((l-m+1)*(l-m+2)/2/(2*l+1)/(2*l+3))*l*ylms[:,l+1,Midx-1]/Ris

            if abs(m-1) <= l-1:
                xcovm1 -= np.sqrt((l+m-1)*(l+m)/2/(2*l-1)/(2*l+1))*(l+1)*ylms[:,l-1,Midx-1]/Ris

            #transform the gradient to cartesian
            gradylms[:,l,midx,0] = 1/np.sqrt(2)*(xcovm1-xcovpl1)
            gradylms[:,l,midx,1] = 1j/np.sqrt(2)*(xcovm1+xcovpl1)
            gradylms[:,l,midx,2] = xcov0

    # index ylms to get rid of extra terms for derivative
    ylms = ylms[:, 0:lmax+1, 1:1+2*lmax+1]
    # multiply the spherical harmonics and the radial inner product
    Y_mul_innerprod = np.einsum('ijk,ilj->iljk', ylms, integral_array)
    # multiply the gradient of the spherical harmonics with the radial inner get_radial_inner_product
    dY_mul_innerprod = np.einsum('ijkn,ilj->iljkn', gradylms, integral_array)
    # multiply the spherical harmonics with the gradient of the radial inner get_radial_inner_product
    Y_mul_dinnerprod = np.einsum('ijk,iljn->iljkn', ylms, grad_integral_array)
    # multiply the gaussians into the expression with 4pi
    C = np.einsum('i,ijkl->ijkl', exparray, Y_mul_innerprod)
    # multiply the gradient of the gaussian with the other terms
    gradexp_mul_y_inner = np.einsum('in,ijkl->ijkln', gradexparray, Y_mul_innerprod)
    # add gradient of inner product and spherical harmonic terms
    gradHarmonics_mul_gaussian = np.einsum('ijkln,i->ijkln', dY_mul_innerprod+Y_mul_dinnerprod, exparray)
    dC = gradexp_mul_y_inner + gradHarmonics_mul_gaussian
    dC *= cutoff_array[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
    dC += np.einsum('in,ijkl->ijkln', grad_cutoff_array, C)
    C *= cutoff_array[:, np.newaxis, np.newaxis, np.newaxis]
    return C, dC

if  __name__ == "__main__":
    from optparse import OptionParser
    from ase.io import read
    import time
    # ---------------------- Options ------------------------
    parser = OptionParser()
    parser.add_option("-c", "--crystal", dest="structure",
                      help="crystal from file, cif or poscar, REQUIRED",
                      metavar="crystal")

    parser.add_option("-r", "--rcut", dest="rcut", default=3.0, type=float,
                      help="cutoff for neighbor calcs, default: 3.0"
                      )

    parser.add_option("-l", "--lmax", dest="lmax", default=2, type=int,
                      help="lmax, default: 2"
                      )

    parser.add_option("-n", "--nmax", dest="nmax", default=2, type=int,
                      help="nmax, default: 2"
                      )

    parser.add_option("-a", "--alpha", dest="alpha", default=2.0, type=float,
                      help="cutoff for neighbor calcs, default: 2.0"
                      )

    parser.add_option("-f", dest="der", default=True,
                      action='store_false',help='derivative flag')

    (options, args) = parser.parse_args()

    if options.structure is None:
        from ase.build import bulk
        test = bulk('Si', 'diamond', a=5.459, cubic=True)
    else:
        test = read(options.structure, format='vasp')

    cell = test.get_cell()
    cell[0,1] += 0.5
    test.set_cell(cell)

    lmax = options.lmax
    nmax = options.nmax
    rcut = options.rcut
    alpha = options.alpha
    der = options.der

    start1 = time.time()
    f = SO3(nmax=nmax, lmax=lmax, rcut=rcut, alpha=alpha)
    start2 = time.time()
    print(f)
    p = f.compute_p(test, atom_ids=[0]); print('from P', p)
    x = f.calculate(test, derivative=der)
    print('x', x['x'])
    dp, p = f.compute_dpdr(test); print('from dP', p)
    dp, p = f.compute_dpdr_5d(test); print('from dP5d', p)

    (x, spg, wps) = ([13.45493847, 0.98, 0.04, 0.86, 0.08, 0.94, 0.14, 0.22, 0.66, 0.22, 0.74, 0.28, 0.2, 0.18, 0.68], 226, ['192j', '192j', '192j', '96i', '192j'])

    from pyxtal import pyxtal
    xtal = pyxtal()
    xtal.from_spg_wps_rep(spg, wps, x, ['C']*len(wps))
    atoms = xtal.to_ase()
    p1 = f.compute_p(atoms)[0]; print("from ase", p1)
    p2 = f.compute_p(atoms, atom_ids=[0])[0]; print("from self", p2)
    print(np.sum((p1-p2)**2))
    assert(np.sum((p1-p2)**2)<1e-4)




# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

import numpy as np
import pymatgen.analysis.structure_matcher as sm
from pymatgen.core import Lattice as pmg_Lattice
from pymatgen.core import Structure
from pymatgen.core.operations import SymmOp

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.operations import get_inverse
from pyxtal.symmetry import Group, Hall, Wyckoff_position


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestNeighbour(unittest.TestCase):
    def test_packing(self):
        c = pyxtal(molecular=True)
        for data in [
            ("aspirin", 14),
            ("WEXBOS", 14),
            ("MERQIM", 12),
            ("LAGNAL", 16),
            ("YICMOP", 14),
            ("LUFHAW", 18),
            ("coumarin", 14),
            ("HAHCOI", 14),
            ("JAPWIH", 14),
            ("AXOSOW01", 14),
            ("PAHYON01", 13),
            ("xxvi", 15),
            ("resorcinol", 14),
        ]:
            (name, CN) = data
            c.from_seed(seed=cif_path + name + ".cif", molecules=[name])
            ds, _, _, _, engs = c.get_neighboring_molecules(0, 1.5)
            # print(engs)
            # print(name, CN, len(ds))
            assert len(ds) == CN


class TestSubgroup(unittest.TestCase):
    def test_cubic_cubic(self):
        sites = ["8a", "32e"]
        numIons = int(sum([int(i[:-1]) for i in sites]))
        C1 = pyxtal()
        C1.from_random(3, 227, ["C"], [numIons], sites=[sites])
        pmg_s1 = C1.to_pymatgen()
        # sga1 = SpacegroupAnalyzer(pmg_s1).get_space_group_symbol()

        C2s = C1.subgroup(eps=1e-5)
        for C2 in C2s:
            pmg_s2 = C2.to_pymatgen()
            # sga2 = SpacegroupAnalyzer(pmg_s2).get_space_group_symbol()
            # prevent some numerical error
            if not sm.StructureMatcher().fit(pmg_s1, pmg_s2):
                C12 = pyxtal()
                C12.from_seed(pmg_s2)
                pmg_12 = C12.to_pymatgen()
                assert sm.StructureMatcher().fit(pmg_s1, pmg_12)

            # self.assertTrue(sm.StructureMatcher().fit(pmg_s1, pmg_s2))

        C1.subgroup(perms={"C": "Si"}, H=216)

    def test_from_seed(self):
        coords = [[0, 0, 0], [0.75, 0.5, 0.75]]
        lattice = pmg_Lattice.from_parameters(a=3.84, b=3.84, c=3.84, alpha=120, beta=90, gamma=60)
        struct = Structure(lattice, ["Si", "C"], coords)
        s1 = pyxtal()
        s1.from_seed(struct)
        s2 = s1.subgroup_once(eps=0)
        pmg_s1 = s1.to_pymatgen()
        pmg_s2 = s2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg_s1, pmg_s2)

    def test_molecules(self):
        for name in [
            "aspirin",
            "resorcinol",
            "coumarin",
            "HAHCOI",
            "xxvi",
            "WEXBOS",
            "MERQIM",
            "LAGNAL",
            "YICMOP",
            "LUFHAW",
            "JAPWIH",
            "AXOSOW01",
            "PAHYON01",
        ]:
            cif = cif_path + name + ".cif"
            struc = pyxtal(molecular=True)
            struc.from_seed(seed=cif, molecules=[name])
            pmg_struc = struc.to_pymatgen()
            pmg_s1 = Structure.from_file(cif, primitive=True)
            assert sm.StructureMatcher().fit(pmg_struc, pmg_s1)

            Cs = struc.subgroup(eps=0, max_cell=1)
            for C in Cs:
                pmg_s2 = C.to_pymatgen()
                assert sm.StructureMatcher().fit(pmg_struc, pmg_s2)

    def test_hydrate(self):
        # glycine dihydrate
        cif = cif_path + "gdh.cif"
        struc = pyxtal(molecular=True)
        struc.from_seed(seed=cif, molecules=["Glycine-z", "H2O"])
        pmg_struc = struc.to_pymatgen()
        pmg_s1 = Structure.from_file(cif, primitive=True)
        assert sm.StructureMatcher().fit(pmg_struc, pmg_s1)

    def test_special(self):
        cif = cif_path + "191.vasp"
        struc = pyxtal()
        struc.from_seed(seed=cif)
        for _i in range(100):
            struc.subgroup_once(0.2, None, None, "t+k", 2)


class TestLoad(unittest.TestCase):
    def test_atomic(self):
        s1 = pyxtal()
        s1.from_random(3, 36, ["C", "Si"], [4, 8])
        s2 = pyxtal()
        s2.load_dict(s1.save_dict())
        pmg_s1 = s1.to_pymatgen()
        pmg_s2 = s2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg_s1, pmg_s2)

    def test_molecular(self):
        s1 = pyxtal(molecular=True)
        s1.from_random(3, 36, ["H2O"], [4])
        s2 = pyxtal()
        s2.load_dict(s1.save_dict())
        pmg_s1 = s1.to_pymatgen()
        pmg_s2 = s2.to_pymatgen()
        assert sm.StructureMatcher().fit(pmg_s1, pmg_s2)


class TestPartial(unittest.TestCase):
    def test_Al2SiO5(self):
        cell = Lattice.from_para(7.8758, 7.9794, 5.6139, 90, 90, 90)
        spg = 58
        elements = ["Al", "Si", "O"]
        composition = [8, 4, 20]
        sites = [
            {
                "4e": [0.0000, 0.0000, 0.2418],
                "4g": [0.1294, 0.6392, 0.0000],
            },
            {"4g": [0.2458, 0.2522, 0.0000]},
            [],  # empty for oxygen
        ]

        s = pyxtal()
        s.from_random(3, spg, elements, composition, lattice=cell, sites=sites)
        assert s.valid

        sites2 = [
            {
                "4e": [0.0000, 0.0000, 0.2418],
                "4g": [0.1294, 0.6392, 0.0000],
            },
            {"4g": [0.2458, 0.2522, 0.0000]},
            {"4g": [0.4241, 0.3636, 0.0000]},  # partial info on O
        ]

        s = pyxtal()
        s.from_random(3, spg, elements, composition, lattice=cell, sites=sites2)
        assert s.valid


class Test_resort(unittest.TestCase):
    def test_molecule(self):
        rng = np.random.default_rng(0)
        # glycine dihydrate
        cif = cif_path + "gdh.cif"
        struc = pyxtal(molecular=True)
        struc.from_seed(seed=cif, molecules=["Glycine-z", "H2O"])
        N1 = len(struc.mol_sites)
        l = list(range(len(struc.mol_sites)))
        rng.shuffle(l)
        struc.mol_sites = [struc.mol_sites[i] for i in l]
        struc.resort()
        N2 = len(struc.mol_sites)
        assert N1 == N2

    def test_atom(self):
        cif = cif_path + "aspirin.cif"
        struc = pyxtal()
        rng = np.random.default_rng(0)
        struc.from_seed(seed=cif)
        N1 = len(struc.atom_sites)
        l = list(range(len(struc.atom_sites)))
        rng.shuffle(l)
        struc.atom_sites = [struc.atom_sites[i] for i in l]
        struc.resort()
        N2 = len(struc.atom_sites)
        assert N1 == N2


class Test_rng(unittest.TestCase):
    """
    Test rng generators in two ways
    1. random_state as a fixed integer
    2. random_state as a generator
    """
    def test_rng_integer(self):
        xtal = pyxtal(); xtal.from_random(3, 194, ['C'], [8], random_state=0)
        xs = xtal.get_1d_rep_x()
        assert np.sum((xs - np.array([4.679, 6.418, 0.943])**2)) < 1e-2

        xtal = pyxtal(molecular=True)
        xtal.from_random(3, 19, ['aspirin'], [4], random_state=0)
        rep = xtal.get_1D_representation().x
        d1 = np.array([115, 17.294, 15.077, 9.018])
        d2 = np.array([0.677, 0.243, 0.612, 0.057, -1.194, 0.110])
        assert np.sum((rep[0] - d1)**2) < 1e-2
        assert np.sum((rep[1][1:-1] - d2)**2) < 1e-2


    def test_rng_generator(self):
        rng = np.random.default_rng(1)
        xtal = pyxtal()
        xtal.from_random(3, 194, ['C'], [8], random_state=rng)
        xs = xtal.get_1d_rep_x()
        assert np.sum((xs - np.array([7.442, 2.110])**2)) < 1e-2

        xtal.from_random(3, 194, ['C'], [8], random_state=rng)
        xs = xtal.get_1d_rep_x()
        assert np.sum((xs - np.array([5.864, 5.809])**2)) < 1e-2

        xtal = pyxtal(molecular=True)
        xtal.from_random(3, 19, ['aspirin'], [4], random_state=rng)
        rep = xtal.get_1D_representation().x
        d1 = np.array([115, 14.207, 18.334, 9.028])
        d2 = np.array([0.294, 0.627, 0.528, 157.052, -11.968, -171.851])
        assert np.sum((rep[0] - d1)**2) < 1e-2
        assert np.sum((rep[1][1:-1] - d2)**2) < 1e-2

        xtal.from_random(3, 19, ['aspirin'], [4], random_state=rng)
        rep = xtal.get_1D_representation().x
        d1 = np.array([115, 12.763, 16.639, 11.073])
        d2 = np.array([0.504, 0.127, 0.585, -21.523, -68.406, 152.839])
        assert np.sum((rep[0] - d1)**2) < 1e-2
        assert np.sum((rep[1][1:-1] - d2)**2) < 1e-2

class Test_operations(unittest.TestCase):
    def test_inverse(self):
        coord0 = [0.35, 0.1, 0.4]
        coords = np.array(
            [
                [0.350, 0.100, 0.400],
                [0.350, 0.100, 0.000],
                [0.350, 0.100, 0.000],
                [0.350, 0.000, 0.667],
                [0.350, 0.000, 0.250],
                [0.350, 0.350, 0.400],
                [0.350, 0.350, 0.500],
                [0.350, 0.350, 0.000],
                [0.350, 0.350, 0.350],
                [0.100, 0.100, 0.100],
                [0.400, 0.400, 0.400],
                [0.350, 0.000, 0.000],
                [0.000, 0.100, 0.400],
                [0.350, 0.000, 0.400],
            ]
        )
        xyzs = [
            "x,y,z",
            "x,y,0",
            "y,x,0",
            "x,0,2/3",
            "0,x,1/4",
            "x,x,z",
            "x,-x,1/2",
            "2x,x,0",
            "-2x,-0.5x,-x+1/4",
            "-2y,-0.5y,-y+1/4",
            "-2z,-0.5z,-z+1/4",
            "0,0,x",
            "-y/2+1/2,-z,0",
            "-z,-x/2+1/2,0",
        ]

        for i, xyz in enumerate(xyzs):
            op = SymmOp.from_xyz_str(xyz)
            inv_op = get_inverse(op)
            coord1 = op.operate(coord0)
            coord2 = inv_op.operate(coord1)
            assert np.allclose(coord2, coords[i], rtol=0.01)
            # strs = "{:6.3f} {:6.3f} {:6.3f}".format(*coord0)
            # strs += "  {:12s}  ".format(op.as_xyz_str())
            # strs += "{:6.3f} {:6.3f} {:6.3f}".format(*coord1)
            # strs += "  {:12s}  ".format(inv_op.as_xyz_str())
            # strs += "{:6.3f} {:6.3f} {:6.3f}".format(*coord2)
            # print(strs)

    def test_swap_wp(self):
        g = Group(38)
        wp = g[4]
        wp1, trans = wp.swap_axis([1, 0, 2])

        g = Group(71)
        wp = g[5]
        wp1, trans = wp.swap_axis([0, 2, 1])
        wp1, trans = wp.swap_axis([1, 2, 0])
        wp1, trans = wp.swap_axis([2, 1, 0])

    def test_alternative(self):
        for name in ["BTO-Amm2", "lt_quartz", "GeF2", "lt_cristobalite", "PVO"]:
            s = pyxtal()
            s.from_seed(cif_path + name + ".cif")
            pmg_s1 = s.to_pymatgen()
            strucs = s.get_alternatives()
            for struc in strucs:
                pmg_s2 = struc.to_pymatgen()
                assert sm.StructureMatcher().fit(pmg_s1, pmg_s2)

    def test_wyc_sets(self):
        for i in range(1, 229):
            Group(i, quick=True).get_alternatives()["No."]

    def test_trans(self):
        mats = [
            np.array([[1, 0, 1], [0, 1, 0], [0, 0, 1]]),
            np.array([[1, 0, -1], [0, 1, 0], [0, 0, 1]]),
        ]

        for n in range(3, 16):
            hns = Hall(n).hall_numbers
            for hn in hns:
                g = Group(hn, use_hall=True)
                for i in range(1, len(g)):
                    wp = g[i].copy()
                    for mat in mats:
                        wp.transform_from_matrix(mat)
                        c1, c2 = wp.update()
                        # print(hn, i, (c1 or c2))
                        assert c1 or c2

    # def test_image(self):
    #    from pyxtal.descriptor import spherical_image
    #    c1 = pyxtal(molecular=True)
    #    for name in ['benzene', 'aspirin', 'naphthalene']:
    #        c1.from_seed(seed=cif_path+name+".cif", molecules=[name])
    #        for model in ['contact', 'molecule']:
    #            sph = spherical_image(c1, model=model)
    #            sph.align()
    #            print(name, model)

    class TestSubstitution(unittest.TestCase):
        def test_substitute_1_2(self):
            data = [
                (227, ["8a"], [3.6], ["C"], 1),
                (
                    92,
                    ["4a", "8b"],
                    [5.0847, 7.0986, 0.2944, 0.0941, 0.2410, 0.8256],
                    ["Si", "O"],
                    1,
                ),
            ]
            for d in data:
                (spg, wps, rep, elements, N) = d
                s = pyxtal()
                xtals = s.from_spg_wps_rep(spg, wps, rep, elements)
                assert len(xtals) == N

        def test_criteria(self):
            criteria = {"CN": {"B": 4, "N": 4}, "cutoff": 1.9, "exclude_ii": True}
            xtals = pyxtal().substitute_1_2({"C": ["B", "N"]}, ratio=[1, 1], criteria=criteria)
            assert xtals[0].check_validity(criteria)


if __name__ == "__main__":
    unittest.main()
